""" Module to create a data iterator for streams and annotations

Author:
    Dominik Schiller <dominik.schiller@uni-a.de>
Date:
    18.8.2023

"""
import numpy as np
import sys
import os
import warnings

from typing import Union
from nova_utils.data.data import Data
from nova_utils.data.stream import Stream, StreamMetaData
from pathlib import Path
from nova_utils.data.handler.mongo_handler import (
    AnnotationHandler,
    StreamHandler,
    SessionHandler,
)
from nova_utils.data.handler.file_handler import FileHandler
from nova_utils.utils import string_utils
from nova_utils.utils.anno_utils import data_contains_garbage
from nova_utils.data.session import Session


class NovaIterator:
    """Iterator class for processing data samples from the Nova dataset.

    The NovaIterator takes all information about what data should be loaded and how it should be processed. The class itself then takes care of loading all data and provides an iterator to directly apply a sliding window to the requested data.
    Every time based argument can be passed either as string or a numerical value. If the time is passed as string, the string should end with either 's' to indicate the time is specified in seconds or 'ms' for milliseconds.
    If the time is passed as a numerical value or as a string without indicating a specific unit it is assumed texthat an integer value represents milliseconds while a float represents seconds. All numbers will be represented as integer milliseconds internally.
    The highest time resolution for processing is therefore 1ms.

    Args:
        db_host (str): Database IP address.
        db_port (int): Database port.
        db_user (str): Database username.
        db_password (str): Database password.
        dataset (str): Name of the dataset.
        data_dir (Path, optional): Path to the data directory. Defaults to None.
        sessions (list[str], optional): List of session names to process. Defaults to None.
        data (list[dict[str, str]], optional): List of data descriptions. Defaults to None. The dictionary should have the following fields:

            ``"id"``:
                Unique id to map the data to a given input / output.
            ``"name"``:
                Output name for streams
            ``"type"``:
                IO type of the data. Either "input" or "output"
            ``"src"``
                The source and datatype to load the data from separated by ':' . Source can be either 'db' for database or 'file' to load from disc.
                The dataytpye is eiter 'stream' or 'anno'. E.g. 'db:anno' .
            ``"scheme"``
                The scheme name of the annotations to load. Only necessary when loading annotations from the database.
            ``"annotator"``
                The annotator of the annotations to load. Only necessary when loading annotations from the database.
            ``"role"``
                The role to which the data belongs. Only necessary when accessing data from the database.
            ``"fp"``
                The filepath from which to load the data from. Only necessary when loading files from disk.

        frame_size (Union[int, float, str], optional): Size of the data frame measured in time. Defaults to None.
        start (Union[int, float, str], optional): Start time for processing measured in time. Defaults to None.
        end (Union[int, float, str], optional): End time for processing measured in time. Defaults to None.
        left_context (Union[int, float, str], optional): Left context duration measured in time. Defaults to None.
        right_context (Union[int, float, str], optional): Right context duration measured in time. Defaults to None.
        stride (Union[int, float, str], optional): Stride for iterating over data measured in time. If stride is not set explicitly it will be set to frame_size. Defaults to None.
        add_rest_class (bool, optional): Whether to add a rest class for discrete annotations. Defaults to True.
        fill_missing_data (bool, optional): Whether to fill missing data. Defaults to True. THIS OPTION IS CURRENTLY NOT DOING ANYTHING

    Attributes:
        data_dir (Path): Path to the data directory.
        dataset (str): Name of the dataset.
        sessions (list[str]): List of session names to process.
        data (list[dict]): List of data descriptions.
        frame_size (int, float, str): Size of the data frame measured in time.
        start (int, float, str): Start time for processing measured in time.
        end (int, float, str): End time for processing measured in time.
        left_context (int, float, str): Left context duration measured in time.
        right_context (int, float, str): Right context duration measured in time.
        stride (int, float, str): Stride for iterating  data measured in time.
        add_rest_class (bool): Whether to add a rest class for discrete annotations.
        fill_missing_data (bool): Whether to fill missing data.

    Example:
        .. code-block:: python

            from dotenv import load_dotenv

            # Load environment variables
            IP = 127.0.0.1
            PORT = 1337
            USER = my_user
            PASSWORD = my_password
            DATA_DIR = /data

            # Define dataset and sessions
            dataset = "test_dataset"
            sessions = ["test_session"]

            # Define data descriptions
            annotation = {
                "id" : "my_transcript_1",
                "type": "input",
                "src": "db:anno",
                "scheme": "transcript",
                "annotator": "test_annotator",
                "role": "test_role",
            }

            stream = {
                "id" : "model_output",
                "type" : "input",
                "src": "db:stream",
                "role": "test_role",
                "name": "extracted_features",
            }

            file = {
                "id": "just_a_file",
                "type": "output",
                "src": "file:stream",
                "fp": "/path/to/my/video/test_video.mp4",
            }

            # Create a NovaIterator instance
            nova_iterator = NovaIterator(
                IP,
                PORT,
                USER,
                PASSWORD,
                dataset,
                DATA_DIR,
                sessions=sessions,
                data=[annotation, file],
                frame_size="5s",
                end="20s",
            )

            # Example: Get the next data sample and set a breakpoint
            a = next(nova_iterator)
    """

    def __init__(
        self,
        # Database connection
        db_host: str,
        db_port: int,
        db_user: str,
        db_password: str,
        dataset: str,
        data_dir: Path = None,
        # Data
        sessions: list[str] = None,
        data: list[dict] = None,
        # Iterator Window
        frame_size: Union[int, float, str] = None,
        start: Union[int, float, str] = None,
        end: Union[int, float, str] = None,
        left_context: Union[int, float, str] = None,
        right_context: Union[int, float, str] = None,
        stride: Union[int, float, str] = None,
        # Iterator properties
        add_rest_class: bool = True,
        fill_missing_data=True,
    ):
        self.data_dir = data_dir
        self.dataset = dataset
        self.sessions = sessions
        self.data = data

        # If stride has not been explicitly set it's the same as the frame size
        if stride is None:
            stride = frame_size

        # Parse all times to milliseconds
        self.left_context = string_utils.parse_time_string_to_ms(left_context)
        self.right_context = string_utils.parse_time_string_to_ms(right_context)
        self.frame_size = string_utils.parse_time_string_to_ms(frame_size)
        self.stride = string_utils.parse_time_string_to_ms(stride)
        self.start = string_utils.parse_time_string_to_ms(start)
        self.end = string_utils.parse_time_string_to_ms(end)

        # Frame size 0 or None indicates that the whole session should be returned as one sample
        if self.frame_size == 0:
            warnings.warn(
                "Frame size should be bigger than zero. Returning whole session as sample."
            )

        # If the end time has not been set we initialize it with sys.maxsize
        if self.end is None or self.end == 0:
            self.end = sys.maxsize

        self.add_rest_class = add_rest_class
        self.fill_missing_data = fill_missing_data
        self.current_session = None

        # Data handler
        self._db_session_handler = SessionHandler(
            db_host, db_port, db_user, db_password
        )
        self._db_anno_handler = AnnotationHandler(
            db_host, db_port, db_user, db_password
        )
        self._db_stream_handler = StreamHandler(
            db_host, db_port, db_user, db_password, data_dir=data_dir
        )
        self._file_handler = FileHandler()

        self._iterable = self._yield_sample()

    def _init_data_from_description(self, data_desc: dict, dataset: str, session: str, header_only: bool = False) -> Data:
        """
        Initialize data from a data description.

        Args:
            data_desc (dict): Data description dictionary.
            dataset (str): Dataset name.
            session (str): Session name.
            header_only (bool): If true only the header information for all session data will be loaded.

        Returns:
            Data: Initialized data object.
        """
        src, type_, = data_desc["src"].split(":")
        if src == "db":
            if type_ == "anno":
                return self._db_anno_handler.load(
                    dataset=dataset,
                    session=session,
                    scheme=data_desc["scheme"],
                    annotator=data_desc["annotator"],
                    role=data_desc["role"],
                    header_only=header_only
                )
            elif type_ == "stream":
                try:
                    return self._db_stream_handler.load(
                        dataset=dataset,
                        session=session,
                        name=data_desc["name"],
                        role=data_desc["role"],
                        header_only=header_only
                    )
                except FileNotFoundError as e:
                    # Only raise file not found error if stream is requested as input
                    if not header_only:
                        raise e
                    # Create empty stream file with params
                    else:
                        # Todo differentiate types
                        empty_stream = Stream(None, -1, name=data_desc['name'], role=data_desc["role"], dataset=dataset, session=session)
                        return empty_stream


            else:
                raise ValueError(f"Unknown data type {type_} for data.")
        elif src == "file":
            return self._file_handler.load(fp=Path(data_desc["fp"]))
        else:
            raise ValueError(f"Unknown source type {src} for data.")


    def _data_description_to_string(self, data_desc: dict) -> str:
        """
        Convert data description to a string representation.

        Args:
            data_desc (dict): Data description dictionary.

        Returns:
            str: String representation of the data description.
        """

        id = data_desc.get("id")
        if id is not None:
            return id

        src, type_ = data_desc["src"].split(":")
        delim = "_"
        if src == "db":
            if type_ == "anno":
                return delim.join(
                    [data_desc["scheme"], data_desc["annotator"], data_desc["role"]]
                )
            elif type_ == "stream":
                return delim.join([data_desc["name"], data_desc["role"]])
            else:
                raise ValueError(f"Unknown data type {type_} for data.")
        elif src == "file":
            return delim.join([data_desc["fp"]])
        else:
            raise ValueError(f"Unknown source type {src} for data.")


    def init_session(self, session_name: str) -> Session:
        """
        Initialize a session.

        Args:
            session_name (str): Name of the session to initialize.

        Returns:
            Session: Initialized session object.
        """
        session = self._db_session_handler.load(self.dataset, session_name)

        """Opens all annotations and data readers"""
        input_data = {}
        output_data_templates = {}
        extra_data = {}

        # setting session data
        for data_desc in self.data:
            if data_desc.get('type') == 'input':
                data_initialized = self._init_data_from_description(
                    data_desc, self.dataset, session_name
                )
                data_id = self._data_description_to_string(data_desc)
                input_data[data_id] = data_initialized
            else:
                data_initialized = self._init_data_from_description(
                    data_desc, self.dataset, session_name, header_only=True
                )
                data_id = self._data_description_to_string(data_desc)
                if data_desc.get('type') == 'output':
                    output_data_templates[data_id] = data_initialized
                else:
                    extra_data[data_id] = data_initialized

        session.input_data = input_data
        session.extra_data = extra_data
        session.output_data_templates = output_data_templates

        # update session duration
        min_dur = session.duration if session.duration is not None else sys.maxsize
        for data_initialized in input_data.values():
            if isinstance(data_initialized, Stream):
                meta_data: StreamMetaData = data_initialized.meta_data
                if meta_data.duration is not None:
                    dur = meta_data.duration
                else:
                    dur = len(data_initialized.data) / meta_data.sample_rate * 1000
                if dur < min_dur:
                    min_dur = dur
        session.duration = min_dur

        if session.duration == sys.maxsize:
            raise ValueError(f"Unable to determine duration for session {session.name}")

        return session


    def _yield_sample(self) -> dict[str, np.ndarray]:
        """
        Yield examples.

        Yields:
            dict: Sampled data for the window.

        Example:
            {
                <scheme>_<annotator>_<role> : [[data1], [data2]... ],
                <file_path> : [[data1], [data2]... ]
            }
        """

        # Needed to sort the samples later and assure that the order is the same as in nova.
        sample_counter = 1

        for session in self.sessions:
            # Init all data objects for the session and get necessary meta information
            self.current_session = self.init_session(session)

            # If frame size is zero or less we return the whole data from the whole session in one sample
            if self.frame_size <= 0:
                _frame_size = min(self.current_session.duration, self.end - self.start)
                _stride = _frame_size
            else:
                _frame_size = self.frame_size
                _stride = self.stride

            # Starting position of the first frame in seconds
            cpos = max(self.left_context, self.start)

            # TODO account for stride and framesize being None
            # Generate samples for this session
            while cpos + self.stride + self.right_context <= min(
                self.end, self.current_session.duration
            ):
                frame_start = cpos
                frame_end = cpos + _frame_size
                window_start = frame_start - self.left_context
                window_end = frame_end + self.right_context

                window_info = (
                    session
                    + "_"
                    + str(window_start / 1000)
                    + "_"
                    + str(window_end / 1000)
                )

                # Load data for frame
                data_for_window = {
                    k: v.sample_from_interval(window_start, window_end)
                    for k, v in self.current_session.input_data.items()
                }

                # Performing sanity checks
                garbage_detected = any(
                    [data_contains_garbage(d) for k, d in data_for_window.items()]
                )

                # Incrementing counter
                cpos += _stride
                sample_counter += 1

                if garbage_detected:
                    continue

                yield data_for_window

    def __iter__(self):
        return self._iterable

    def __next__(self):
        return self._iterable.__next__()

    # def get_output_info(self):
    #     def map_label_id(lid):
    #         if self.flatten_samples and not lid == "frame":
    #             return split_role_key(lid)[-1]
    #         return lid
    #
    #     return {
    #         # Adding fake framenumber label for sorting
    #         "frame": {"dtype": np.str, "shape": (1,)},
    #         **{map_label_id(k): v.get_info()[1] for k, v in self.annos.items()},
    #         **{map_label_id(k): v.get_info()[1] for k, v in self.data_info.items()},
    # }


if __name__ == "__main__":
    from dotenv import load_dotenv

    load_dotenv("../../../.env")
    IP = os.getenv("NOVA_IP", "")
    PORT = int(os.getenv("NOVA_PORT", 0))
    USER = os.getenv("NOVA_USER", "")
    PASSWORD = os.getenv("NOVA_PASSWORD", "")
    DATA_DIR = Path(os.getenv("NOVA_DATA_DIR", None))

    dataset = "test"
    sessions = ["04_Oesterreich_test"]

    annotation = {
        "src": "db:anno",
        "scheme": "diarization",
        "type": "input",
        "id": "annotation",
        "annotator": "schildom",
        "role": "testrole2",
    }

    stream = {
        "src": "db:stream",
        "type": "input",
        "id": "featurestream",
        "role": "testrole",
        "name": "arousal.synchrony[testrole2]",
    }

    file = {
        "src": "file:stream",
        "type": "input",
        "id": "file",
        "fp": "/Users/dominikschiller/Work/github/nova-utils/test_files/new_test_video.mp4",
    }

    nova_iterator = NovaIterator(
        IP,
        PORT,
        USER,
        PASSWORD,
        dataset,
        DATA_DIR,
        sessions=sessions,
        data=[annotation],
        frame_size="25s",
        end="100s",
    )

    a = next(nova_iterator)
    breakpoint()



"""Definition of all Annotation classes and Metadata
Author: Dominik Schiller <dominik.schiller@uni-a.de>
Date: 18.8.2023
"""
import numpy as np
import sys
from abc import ABC, abstractmethod
from numpy import dtype
from nova_utils.data.data import DynamicData
from nova_utils.utils.anno_utils import get_overlap, label_is_garbage
from nova_utils.utils.type_definitions import LabelDType, SchemeType


class AnnoMetaData:
    """
    Metadata for annotations, providing information about the annotator.

    Attributes:
        annotator (str, optional): Annotator identifier.
        duration (float, optional): Duration of the stream in ms.

    Args:
        annotator (str, optional): Annotator identifier.
    """

    def __init__(self, annotator: str = None, duration: int = None):
        """
        Initialize an AnnoMetaData instance with annotator information.
        """
        self.annotator = annotator
        self.duration = duration


class IAnnotationScheme(ABC):
    """
    Abstract base class for annotation schemes.

    This class defines the interface for annotation schemes used in the system.

    Attributes:
        name (str): The name of the annotation scheme.

    Args:
        name (str): The name of the annotation scheme.
    """

    def __init__(self, name: str):
        """
        Initialize an IAnnotationScheme instance with the given name.
        """
        self.name = name

    @property
    @abstractmethod
    def scheme_type(self) -> SchemeType:
        """Get the type of the annotation scheme."""
        pass

    @property
    @abstractmethod
    def label_dtype(self) -> dtype:
        """Get the numpy data type of the labels used in the annotation scheme."""
        pass


class DiscreteAnnotationScheme(IAnnotationScheme):
    """
    Discrete annotation scheme class.

    Attributes:
        classes (dict): Dictionary mapping class IDs to class names.

    Args:
        classes (dict): Dictionary mapping class IDs to class names.
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.
    """

    def __init__(self, *args, classes: dict, **kwargs):
        """
        Initialize a DiscreteAnnotationScheme instance with the provided class information.
        """
        super().__init__(*args, **kwargs)
        self.classes = classes

    @classmethod
    @property
    def label_dtype(self) -> dtype:
        """Get the numpy data type of discrete labels."""
        return LabelDType.DISCRETE.value

    @classmethod
    @property
    def scheme_type(self) -> SchemeType:
        """Get the type of the annotation scheme (Discrete)."""
        return SchemeType.DISCRETE


class ContinuousAnnotationScheme(IAnnotationScheme):
    """
    Continuous annotation scheme class.

    Attributes:
        sample_rate (float): Sampling rate of the continuous data.
        min_val (float): Minimum value of the continuous data.
        max_val (float): Maximum value of the continuous data.

    Args:
        sample_rate (float): Sampling rate of the continuous data.
        min_val (float): Minimum value of the continuous data.
        max_val (float): Maximum value of the continuous data.
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.
    """

    def __init__(
        self, *args, sample_rate: float, min_val: float, max_val: float, **kwargs
    ):
        """
        Initialize a ContinuousAnnotationScheme instance with the provided sampling rate and value range.
        """
        super().__init__(*args, **kwargs)
        self.sample_rate = sample_rate
        self.min_val = min_val
        self.max_val = max_val

    @classmethod
    @property
    def label_dtype(self) -> dtype:
        """Get the numpy data type of continuous labels."""
        return LabelDType.CONTINUOUS.value

    @classmethod
    @property
    def scheme_type(self) -> SchemeType:
        """Get the type of the annotation scheme (Continuous)."""
        return SchemeType.CONTINUOUS


class FreeAnnotationScheme(IAnnotationScheme):
    """
    Free annotation scheme class.

    This scheme is used for any form of free text.

    Args:
        name (str): The name of the annotation scheme.
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.
    """

    def __init__(self, *args, **kwargs):
        """
        Initialize a FreeAnnotationScheme instance.
        """
        super().__init__(*args, **kwargs)

    @classmethod
    @property
    def label_dtype(self) -> dtype:
        """Get the numpy data type of free text labels."""
        return LabelDType.FREE.value

    @classmethod
    @property
    def scheme_type(self) -> SchemeType:
        """Get the type of the annotation scheme (Free)."""
        return SchemeType.FREE


class Annotation(DynamicData):
    """
    Base class for annotations.

    Attributes:
        GARBAGE_LABEL_ID (float): Constant representing garbage label.

    Args:
        data (np.ndarray, optional): Array of annotation data.
        scheme (IAnnotationScheme): The annotation scheme used for the data.
        annotator (str, optional): Annotator identifier. Will be added to metadata.
        duration (float, optional): Duration of the annotation in seconds.Will be added to metadata.
        **kwargs: Arbitrary keyword arguments.
    """

    GARBAGE_LABEL_ID = np.NAN

    def __init__(
        self,
        scheme: IAnnotationScheme,
        annotator: str = None,
        duration: float = None,
        **kwargs,
    ):
        """
        Initialize an Annotation instance with data, scheme, and annotator information.
        """
        super().__init__(**kwargs)
        self.annotation_scheme = scheme

        # Create meta data
        anno_meta_data = AnnoMetaData(annotator=annotator, duration=duration)
        self.meta_data.expand(anno_meta_data)

    @property
    @abstractmethod
    def annotation_scheme(self) -> IAnnotationScheme:
        """Get the annotation scheme used for the data."""
        pass

    @annotation_scheme.setter
    def annotation_scheme(self, value):
        pass


class DiscreteAnnotation(Annotation):
    """
    Discrete annotation class.

    Attributes:
        NOVA_REST_CLASS_NAME (int): Constant representing NOVA rest class Name.
        NOVA_GARBAGE_LABEL_ID (int): Constant representing NOVA garbage label.
        rest_label_id (int): Class-variable representing the nova rest class id.

    Args:
        data (np.ndarray): Array of discrete annotation data.
        scheme (DiscreteAnnotationScheme): The discrete annotation scheme used for the data.
        **kwargs: Arbitrary keyword arguments.
    """

    # Class ids and string names as provided from NOVA-DB and required by SSI
    NOVA_REST_CLASS_NAME = "REST"
    NOVA_GARBAGE_LABEL_ID = -1

    def __init__(
        self, scheme: DiscreteAnnotationScheme, data: np.ndarray = None, **kwargs
    ):
        """
        Initialize a DiscreteAnnotation instance with data and scheme information.
        """
        super().__init__(scheme=scheme, **kwargs)
        self._data_values = None
        self._data_interval = None
        self.data = data

    @property
    def annotation_scheme(self) -> DiscreteAnnotationScheme:
        """Get the discrete annotation scheme used for the data."""
        assert isinstance(self._annotation_scheme, DiscreteAnnotationScheme)
        return self._annotation_scheme

    @annotation_scheme.setter
    def annotation_scheme(self, value):
        if not isinstance(value, DiscreteAnnotationScheme):
            raise TypeError(f"Expecting {DiscreteAnnotationScheme}, got {type(value)}.")
        self._annotation_scheme = value
        # Rest class ID is last label ID + 1
        self.rest_label_id = len(self._annotation_scheme.classes)

    @property
    def data(self) -> np.ndarray:
        """Get the annotation data."""
        return self._data

    @data.setter
    def data(self, value):
        if value is not None and not isinstance(value, np.ndarray):
            value = np.asarray(value, dtype=self.annotation_scheme.label_dtype)
        assert value is None or value.dtype == self.annotation_scheme.label_dtype
        self._data = value
        if value is not None:
            # Extract 'from' and 'to' columns and create a new array
            self._data_interval = np.empty((len(value), 2), dtype="<i4")
            self._data_interval[:, 0] = value["from"]
            self._data_interval[:, 1] = value["to"]
            # df_tmp = pd.DataFrame(value)
            # self._data_interval = value[['from','to']]
            # self._data_values = df_tmp[["id", "conf"]].values

    def sample_from_interval(self, start: int, end: int):
        # TODO handle self.data == None
        """
        Sample annotation from an interval.

        Args:
            start (int): Start of the interval in milliseconds.
            end (int): End of the interval milliseconds.

        Returns:
            int: Sampled annotation label.
        """

        overlap_idxs = get_overlap(self._data_interval, start, end)
        dist = np.zeros(
            (len(self.annotation_scheme.classes) + 1),
        )

        # For each sample point where we have an overlap with the label
        for annotation in self.data[overlap_idxs]:
            dur = np.minimum(end, annotation["to"]) - np.maximum(
                start, annotation["from"]
            )

            # If one label is garbage return garbage
            if int(annotation["id"]) is self.NOVA_GARBAGE_LABEL_ID:
                return self.GARBAGE_LABEL_ID

            dist[annotation["id"]] += dur

        # Rest class takes the rest amount of time
        dist[-1] = end - start - sum(dist)

        return dist / sum(dist)


class FreeAnnotation(Annotation):
    """
    Free annotation class.

    Attributes:
        NOVA_GARBAGE_LABEL_VALUE (float): Constant representing NOVA garbage label value.

    Args:
        data (np.ndarray): Array of free text annotation data.
        scheme (FreeAnnotationScheme): The free annotation scheme used for the data.
    """

    NOVA_GARBAGE_LABEL_VALUE = np.NAN

    def __init__(self, scheme: FreeAnnotationScheme, data: np.ndarray = None, **kwargs):
        """
        Initialize a FreeAnnotation instance with data and scheme information.
        """
        super().__init__(scheme=scheme, **kwargs)
        self._data_values = None
        self._data_interval = None
        self.data = data

    @property
    def annotation_scheme(self) -> FreeAnnotationScheme:
        """Get the free annotation scheme used for the data."""
        assert isinstance(self._annotation_scheme, FreeAnnotationScheme)
        return self._annotation_scheme

    @annotation_scheme.setter
    def annotation_scheme(self, value):
        if not isinstance(value, FreeAnnotationScheme):
            raise TypeError(f"Expecting {FreeAnnotationScheme}, got {type(value)}.")
        self._annotation_scheme = value

    @property
    def data(self) -> np.ndarray:
        """Get the annotation data."""
        return self._data

    @data.setter
    def data(self, value):
        if value is not None and not isinstance(value, np.ndarray):
            value = np.asarray(value, dtype=self.annotation_scheme.label_dtype)
        assert value is None or value.dtype == self.annotation_scheme.label_dtype
        self._data = value
        if value is not None:
            # Extract 'from' and 'to' columns and create a new array
            self._data_interval = np.empty((len(value), 2), dtype="<i4")
            self._data_interval[:, 0] = value["from"]
            self._data_interval[:, 1] = value["to"]

            # self._data_interval = value[['from', 'to']]
            # self._data_values = value[['name', 'conf']]
            # df_tmp = pd.DataFrame(value)
            # self._data_interval = df_tmp[["from", "to"]].values
            # self._data_values = df_tmp[["name", "conf"]].values

    def sample_from_interval(self, start: int, end: int):
        """
        Sample annotation from an interval.

        Args:
            start (int): Start of the interval in milliseconds.
            end (int): End of the interval in milliseconds.

        Returns:
            np.ndarray: Numpy array of sampled annotation labels. If no annotations are falling in the specified interval an empty numpy array will be returned.
        """
        annos_for_sample = get_overlap(self._data_interval, start, end)

        if not annos_for_sample.any():
            return np.asarray([])

        return self.data[annos_for_sample]["name"]

    # return self._data_values[annos_for_sample, 0]


class ContinuousAnnotation(Annotation):
    """
    Continuous annotation class.

    Attributes:
        NOVA_GARBAGE_LABEL_VALUE (float): Constant representing NOVA garbage label value.
        MISSING_DATA_LABEL_VALUE (float): Constant representing missing data label value.

    Args:
        data (np.ndarray): Array of continuous annotation data.
        scheme (ContinuousAnnotationScheme): The continuous annotation scheme used for the data.
        **kwargs: Arbitrary keyword arguments.
    """

    NOVA_GARBAGE_LABEL_VALUE = np.NAN
    MISSING_DATA_LABEL_VALUE = sys.float_info.min

    def __init__(
        self, scheme: ContinuousAnnotationScheme, data: np.ndarray = None, **kwargs
    ):
        """
        Initialize a DiscreteAnnotation instance with data and scheme information.
        """
        super().__init__(scheme=scheme, **kwargs)
        self.data = data

    @property
    def annotation_scheme(self) -> ContinuousAnnotationScheme:
        """Get the continuous annotation scheme used for the data."""
        assert isinstance(self._annotation_scheme, ContinuousAnnotationScheme)
        return self._annotation_scheme

    @annotation_scheme.setter
    def annotation_scheme(self, value):
        if not isinstance(value, ContinuousAnnotationScheme):
            raise TypeError(
                f"Expecting {ContinuousAnnotationScheme}, got {type(value)}."
            )
        self._annotation_scheme = value

    @property
    def data(self) -> np.ndarray:
        """Get the annotation data."""
        return self._data

    @data.setter
    def data(self, value):
        if value is not None and not isinstance(value, np.ndarray):
            value = np.asarray(value, dtype=self.annotation_scheme.label_dtype)
        assert value is None or value.dtype == self.annotation_scheme.label_dtype
        self._data = value

    def sample_from_interval(self, start, end):
        """
        Sample annotation from an interval.

        Args:
            start (int): Start of the interval in milliseconds.
            end (int): End of the interval in milliseconds.

        Returns:
            float: Sampled annotation label.
        """
        s = int(start * self.annotation_scheme.sample_rate / 1000)
        e = int(end * self.annotation_scheme.sample_rate / 1000)

        if s == e:
            e = s + 1

        if len(self.data) >= e:
            frame = self.data[s:e]
            frame_data = frame[:, 0]
            frame_conf = frame[:, 1]
        else:
            return self.MISSING_DATA_LABEL_VALUE

        # conf = sum(frame_conf) / max(len(frame_conf), 1)
        label = sum(frame_data) / max(len(frame_data), 1)

        if label_is_garbage(label, self.NOVA_GARBAGE_LABEL_VALUE):
            return self.NOVA_GARBAGE_LABEL_VALUE
        else:
            return label


if __name__ == "__main__":
    # Discrete annotation example
    discrete_scheme = DiscreteAnnotationScheme(
        name="disc_scheme", classes={0: "class_zero", 1: "class_one", 2: "class_two"}
    )
    discrete_data = np.array(
        [
            (500, 1000, 0, 0.8),
            (1500, 2000, 2, 0.6),
            (2500, 3000, 1, 0.9),
        ],
        dtype=discrete_scheme.label_dtype,
    )

    discrete_anno = DiscreteAnnotation(data=discrete_data, scheme=discrete_scheme)

    # Continuous annotation example
    continuous_scheme = ContinuousAnnotationScheme(
        name="continuous_scheme", sample_rate=0.25, min_val=0, max_val=1
    )
    continuous_data = np.array(
        [
            (0.7292248, 0.52415526),
            (0.2252654, 0.4546865),
            (0.64103144, 0.7247994),
            (0.3928702, 0.5221592),
            (0.05887425, 0.58045745),
            (0.19909602, 0.01523399),
            (0.8669538, 0.8970701),
            (0.89999694, 0.80160624),
            (0.33919978, 0.7137072),
            (0.5318645, 0.53093654),
        ],
        dtype=continuous_scheme.label_dtype,
    )
    continuous_anno = ContinuousAnnotation(
        data=continuous_data,
        scheme=continuous_scheme,
    )

    # Free annotation example
    free_scheme = FreeAnnotationScheme(name="free_scheme")
    free_data = np.array(
        [
            (1250, 2750, "hello", 0.75),
            (3140, 5670, "world", 0.82),
            (250, 750, "yehaaaaw", 0.62),
            (7890, 9100, "!!!", 0.91),
        ],
        dtype=free_scheme.label_dtype,
    )
    free_anno = FreeAnnotation(
        data=free_data,
        scheme=free_scheme,
    )



""" Module for handling File data operations related to annotations and streams.

Author:
    Dominik Schiller <dominik.schiller@uni-a.de>
Date:
    18.8.2023

"""

import xml.etree.ElementTree as Et
import numpy as np
import csv
import decord
import subprocess
import json
import math
from typing import Union
from decord import cpu
from struct import *
from nova_utils.data.data import Data
from pathlib import Path
from nova_utils.data.handler.ihandler import IHandler
from nova_utils.data.annotation import (
    SchemeType,
    Annotation,
    DiscreteAnnotation,
    DiscreteAnnotationScheme,
    ContinuousAnnotation,
    ContinuousAnnotationScheme,
    FreeAnnotation,
    FreeAnnotationScheme,
)
from nova_utils.data.stream import (
    Video,
    Audio,
    Stream,
    SSIStream,
    StreamMetaData,
    SSIStreamMetaData,
)
import mmap
import ffmpegio
from nova_utils.utils.type_definitions import LabelDType, SSILabelDType, SSIFileType, SSINPDataType
from nova_utils.utils.anno_utils import convert_label_to_ssi_dtype, convert_ssi_to_label_dtype
from nova_utils.utils.string_utils import string_to_enum

# METADATA
class FileMetaData:
    """Metadata for a file.

    Attributes:
        file_path (Path): The filepath from which the data has been loaded
    """

    def __init__(self, file_path: Path):
        """
        Initialize metadata for any file.

        Args:
        file_path (Path): The filepath from which the data has been loaded
        """
        self.file_path = file_path


class FileSSIStreamMetaData:
    """Metadata for a file.

    Attributes:
        ftype (str): The file type.
        delim (str): The delimiter used to separate entries in the file.
    """

    def __init__(self, ftype: str, delim: str):
        """
        Initialize additional metadata for a file-based SSIStream.
        FileMetaData needs to be set separately.

        Args:
            ftype (str): The file type.
            delim (str): The delimiter used in the file.
        """
        self.ftype = ftype
        self.delim = delim


# ANNOTATIONS

class _AnnotationFileHandler(IHandler):
    """Class for handling the loading and saving of data annotations."""

    @staticmethod
    def _load_data_discrete(path: Path, ftype: str):
        """
        Load discrete annotation data from a file.

        Args:
            path (Path): The path of the file containing the annotation data.
            ftype (str): The file type (ASCII or BINARY) of the annotation data.

        Returns:
            np.ndarray: The loaded discrete annotation data as a NumPy array.
        """

        if ftype == SSIFileType.ASCII.name:
            data =  np.loadtxt(path, dtype=SSILabelDType.DISCRETE.value, delimiter=";")
        elif ftype == SSIFileType.BINARY.name:
            data = np.fromfile(path, dtype=SSILabelDType.DISCRETE.value)
        else:
            raise ValueError("FileType {} not supported".format(ftype))

        return data

    @staticmethod
    def _load_data_continuous(path, ftype):
        """
        Load continuous annotation data from a file.

        Args:
            path (Path): The path of the file containing the annotation data.
            ftype (str): The file type (ASCII or BINARY) of the annotation data.

        Returns:
            np.ndarray: The loaded continuous annotation data as a NumPy array.
        """
        if ftype == SSIFileType.ASCII.name:
            data = np.loadtxt(path, dtype=SSILabelDType.CONTINUOUS.value, delimiter=";")
        elif ftype == SSIFileType.BINARY.name:
            data = np.fromfile(path, dtype=SSILabelDType.CONTINUOUS.value)
        else:
            raise ValueError("FileType {} not supported".format(ftype))

        return data
    @staticmethod
    def _load_data_free(path, ftype, size):
        """
        Load free annotation data from a file.

        Args:
            path (Path): The path of the file containing the annotation data.
            ftype (str): The file type (ASCII or BINARY) of the annotation data.
            size (int): The size of the data to be loaded.

        Returns:
            np.ndarray: The loaded free annotation data as a NumPy array.
        """
        data = []
        if ftype == SSIFileType.ASCII.name:
            with open(path, "r") as ascii_file:
                ascii_file_reader = csv.reader(ascii_file, delimiter=";", quotechar='"')
                for row in ascii_file_reader:
                    f = float(row[0])
                    t = float(row[1])
                    n = row[2]
                    c = float(row[3])
                    data.append((f, t, n, c))

        elif ftype == SSIFileType.BINARY.name:
            with open(path, "rb") as binary_file:
                counter = 0
                binary_file.seek(0)

                while counter < size:
                    # from (8byte float)
                    f = unpack("d", binary_file.read(8))[0]
                    # to (8byte float)
                    t = unpack("d", binary_file.read(8))[0]
                    # length of label (4byte uint)
                    lol = unpack("i", binary_file.read(4))[0]
                    # the label (lol * byte)
                    n = binary_file.read(lol).decode("ISO-8859-1")
                    # confidence (4Byte float)
                    c = unpack("f", binary_file.read(4))[0]

                    data.append((f, t, n, c))
                    counter += 1
        else:
            raise ValueError("FileType {} not supported".format(ftype))

        return np.asarray(data, dtype=SSILabelDType.FREE.value)

    @staticmethod
    def _str_format_from_dtype(dtype: np.dtype):
        """
        Generate a string format for a given numpy dtype.

        Args:
            dtype (np.dtype): The numpy dtype.

        Returns:
            list: A list of format strings for each field in the dtype.
        """
        fmt = []

        for _, field_info in dtype.fields.items():
            dt, bo = field_info
            if np.issubdtype(dt, np.integer):
                # For integers, use '%d' format
                format_string = "%d"
            elif np.issubdtype(dt, np.floating):
                # For floating-point numbers, use '%.2f' format with 2 decimal places
                format_string = "%.2f"
            else:
                # For other data types (e.g., strings, bools, etc.), use the default '%s' format
                format_string = "%s"
            fmt.append(format_string)

        return fmt

    def load(self, fp: Path, header_only: bool = False) -> Annotation:
        """
        Load annotation data from an XML file.

        Args:
            fp (Path): The file path of the XML annotation file.
            header_only (bool): If true only the stream header will be loaded.


        Returns:
            Annotation: The loaded annotation data as an Annotation object.
        """
        data_path = fp.with_suffix(fp.suffix + "~")
        tree = Et.parse(fp)

        # info
        info = tree.find("info", {})
        ftype = info.get("ftype")
        size = int(info.get("size", 0))

        # meta
        meta = tree.find("meta", {})
        role = meta.get("role")
        annotator = meta.get("annotator")

        # scheme
        scheme = tree.find("scheme", {})
        scheme_name = scheme.get("name")
        scheme_type = scheme.get("type")

        # TODO: Nova Annotations do export a 'color' column where ssi annotations do not. Account for this
        # anno object
        # discrete scheme
        if scheme_type == SchemeType.DISCRETE.name:
            scheme_classes = {}
            for item in scheme:
                scheme_classes[item.get("id")] = item.get("name")

            anno_data = None
            if not header_only:
                anno_data = self._load_data_discrete(data_path, ftype)
                anno_data = convert_ssi_to_label_dtype(anno_data, SchemeType.DISCRETE)

            anno_scheme = DiscreteAnnotationScheme(
                name=scheme_name, classes=scheme_classes
            )
            annotation = DiscreteAnnotation(
                data=anno_data,
                scheme=anno_scheme,
                role=role,
                annotator=annotator,
            )

        # continuous scheme
        elif scheme_type == SchemeType.CONTINUOUS.name:
            sr = float(scheme.get("sr"))
            min_val = float(scheme.get("min"))
            max_val = float(scheme.get("max"))

            anno_data = None
            if not header_only:
                anno_data = self._load_data_continuous(data_path, ftype)
                anno_data = convert_ssi_to_label_dtype(anno_data, SchemeType.CONTINUOUS)
            anno_scheme = ContinuousAnnotationScheme(
                name=scheme_name, sample_rate=sr, min_val=min_val, max_val=max_val
            )
            annotation = ContinuousAnnotation(
                scheme=anno_scheme,
                data=anno_data,
                role=role,
                annotator=annotator,
            )

        # free scheme
        elif scheme_type == SchemeType.FREE.name:
            anno_data = None
            if not header_only:
                anno_data = self._load_data_free(data_path, ftype, size)
                anno_data = convert_ssi_to_label_dtype(anno_data, SchemeType.FREE)
            anno_scheme = FreeAnnotationScheme(name=scheme_name)
            annotation = FreeAnnotation(
                scheme=anno_scheme,
                data=anno_data,
                role=role,
                annotator=annotator,
            )
        else:
            raise TypeError(f"Unknown scheme type {type}")

        return annotation

    def save(self, data: Annotation, fp: Path, ftype: SSIFileType = SSIFileType.ASCII):
        """
        Save annotation data to a file.

        Args:
            data (Annotation): The annotation data to be saved.
            fp (Path): The file path for saving the data.
            ftype (SSIFileTypes, optional): The file type (ASCII or BINARY) for saving.

        Raises:
            TypeError: If filetype is not supported for saving or unknown
        """
        data_path = fp.with_suffix(fp.suffix + "~")

        # header
        root = Et.Element("annotation", attrib={"ssi-v ": "3"})

        # info
        size = str(len(data.data))
        Et.SubElement(root, "info", attrib={"ftype": ftype.name, "size": size})

        # meta
        role = "" if data.meta_data.role is None else data.meta_data.role
        annotator = "" if data.meta_data.annotator is None else data.meta_data.annotator
        Et.SubElement(root, "meta", attrib={"role": role, "annotator": annotator})

        # scheme
        scheme_name = data.annotation_scheme.name
        scheme_type = data.annotation_scheme.scheme_type

        if scheme_type == SchemeType.DISCRETE:
            data: DiscreteAnnotation
            scheme = Et.SubElement(
                root, "scheme", attrib={"name": scheme_name, "type": scheme_type.name}
            )
            for class_id, class_name in data.annotation_scheme.classes.items():
                Et.SubElement(
                    scheme, "item", attrib={"name": class_name, "id": class_id}
                )

        elif scheme_type == SchemeType.CONTINUOUS:
            data: ContinuousAnnotation
            Et.SubElement(
                root,
                "scheme",
                attrib={
                    "name": scheme_name,
                    "type": scheme_type.name,
                    "sr": f"{data.annotation_scheme.sample_rate:.3f}",
                    "min": f"{data.annotation_scheme.min_val:.3f}",
                    "max": f"{data.annotation_scheme.max_val:.3f}",
                },
            )

        elif scheme_type == SchemeType.FREE:
            if ftype == SSIFileType.BINARY:
                raise TypeError(
                    "Binary output format is not supported for free annotation schemes"
                )
            data: FreeAnnotation
            Et.SubElement(
                root, "scheme", attrib={"name": scheme_name, "type": scheme_type.name}
            )
        else:
            raise TypeError(f"Unknown scheme type {type}")

        root = Et.ElementTree(root)
        Et.indent(root, space="    ", level=0)
        root.write(fp)

        anno_data = convert_label_to_ssi_dtype(data.data, scheme_type)

        # save data
        if ftype == SSIFileType.ASCII:
            fmt = self._str_format_from_dtype(anno_data.dtype)
            np.savetxt(data_path, anno_data, fmt=fmt, delimiter=";")
        if ftype == SSIFileType.BINARY:
            data.data.tofile(data_path, sep="")


# SSI STREAMS
class _SSIStreamFileHandler(IHandler):
    """Class for handling the loading and saving of SSIStreams."""

    def _load_header(self, fp: Path) -> dict:
        """
        Load SSIStream header from a file.

        Args:
            fp (Path): The file path of the SSIStream.

        Returns:
            dict: A dictionary containing SSIStream header data.
        """

        tree = Et.parse(fp)

        # info
        info = tree.find("info", {})
        ftype = info.get("ftype")
        sr = info.get("sr")
        dim = info.get("dim")
        byte = info.get("byte")
        dtype = info.get("type")
        delim = info.get("delim")

        # chunks
        chunks = []
        for chunk in tree.findall("chunk"):
            from_ = chunk.get("from")
            to_ = chunk.get("to")
            byte_ = chunk.get("byte")
            num_ = chunk.get("num")
            chunks.append((from_, to_, byte_, num_))

        chunks = np.array(chunks, dtype=SSIStream.CHUNK_DTYPE)
        num_samples = int(sum(chunks["num"]))
        duration = int(num_samples / float(sr) * 1000)

        ssistream_meta_data = {
            "name": fp.stem,
            "ext": fp.suffix,
            "duration": duration,
            "sample_shape": (int(dim),),
            "num_samples": num_samples,
            "sample_rate": float(sr),
            "dtype": string_to_enum(SSINPDataType, dtype).value,
            "chunks": chunks,
            "fp": fp,
            "delim": delim,
            "ftype": ftype,
        }
        return ssistream_meta_data

    def _load_data(
        self,
        fp: Path,
        size: int,
        dim: int,
        ftype=SSIFileType.ASCII,
        dtype: np.dtype = SSINPDataType.FLOAT.value,
        delim=" ",
    ):
        """
        Load SSIStream data from a file.

        Args:
            fp (Path): The file path of the SSIStream data.
            size (int): The size of the data.
            dim (int): The dimension of the data.
            ftype (SSIFileTypes, optional): The file type (ASCII or BINARY) of the SSIStream data.
            dtype (np.dtype, optional): The data type.
            delim (str, optional): The delimiter used in the file.

        Returns:
            np.ndarray: The loaded SSIStream data as a NumPy array.

        Raises:
            ValueError: If the provided filetype is not supported
        """
        if ftype == SSIFileType.ASCII:
            return np.loadtxt(fp, dtype=dtype, delimiter=delim)
        elif ftype == SSIFileType.BINARY:
            return np.fromfile(fp, dtype=dtype).reshape(size, dim)
        else:
            raise ValueError("FileType {} not supported".format(self))

    def save(
        self,
        data: SSIStream,
        fp: Path,
        ftype: SSIFileType = SSIFileType.BINARY,
        delim: str = " ",
    ):
        """
        Save SSIStream data to a file.

        Args:
            data (SSIStream): The SSIStream data to be saved.
            fp (Path): The file path for saving the data.
            ftype (SSIFileTypes, optional): The file type (ASCII or BINARY) for saving.
            delim (str, optional): The delimiter to be used in the file.
        """
        # save header
        data_path = fp.with_suffix(fp.suffix + "~")

        # header
        root = Et.Element("annotation", attrib={"ssi-v ": "2"})

        # info
        meta_data: StreamMetaData | SSIStreamMetaData = data.meta_data
        sr = meta_data.sample_rate
        dim = meta_data.sample_shape[0] if not meta_data.sample_shape is None else data.data[0].shape[0]
        byte = np.dtype(meta_data.dtype).itemsize
        dtype = SSINPDataType(meta_data.dtype).name
        Et.SubElement(
            root,
            "info",
            attrib={
                "ftype": ftype.name,
                "sr": f"{sr:.3f}",
                "dim": str(dim),
                "byte": str(byte),
                "type": dtype,
                "delim": delim,
            },
        )

        # meta
        Et.SubElement(root, "meta")

        # chunks
        for chunk in meta_data.chunks:
            Et.SubElement(
                root,
                "chunk",
                attrib={
                    "fromm": f"{chunk['from']:.3f}",
                    "to": f"{chunk['to']:.3f}",
                    "byte": str(chunk["byte"]),
                    "num": str(chunk["num"]),
                },
            )

        # saving
        root = Et.ElementTree(root)
        Et.indent(root, space="    ", level=0)
        root.write(fp)

        # save data
        if ftype == SSIFileType.ASCII:
            np.savetxt(data_path, data.data, delimiter=delim)
        if ftype == SSIFileType.BINARY:
            data.data.tofile(data_path)

    def load(self, fp, header_only: bool = False, **kwargs) -> Data:
        """
        Load SSIStream data from a file.

        Args:
            fp (Path): The file path of the SSIStream.
            header_only (bool): If true only the stream header will be loaded.

            **kwargs: Arbitrary keyword arguments.

        Returns:
            Data: The loaded SSIStream data.
        """
        data_path = fp.with_suffix(fp.suffix + "~")
        header = self._load_header(fp)
        duration = header.get("duration")
        sample_shape = header.get("sample_shape")
        num_samples = header.get("num_samples")
        sample_rate = header.get("sample_rate")
        dtype = header.get("dtype")
        chunks = header.get("chunks")
        delim = header["delim"]
        ftype = header["ftype"]
        name = header['name']
        ext = header['ext']

        data = None
        if not header_only:
            data = self._load_data(
                fp=data_path,
                size=num_samples,
                dtype=dtype,
                dim=sample_shape[0],
                delim=delim,
                ftype=SSIFileType[ftype],
            )

        ssi_stream = SSIStream(
            data=data,
            duration=duration,
            sample_shape=sample_shape,
            num_samples=num_samples,
            sample_rate=sample_rate,
            dtype=dtype,
            chunks=chunks,
            name=name,
            ext=ext
        )
        ssi_stream.meta_data.expand(FileSSIStreamMetaData(delim=delim, ftype=ftype))
        return ssi_stream


# VIDEO
class _LazyArray(np.ndarray):
    """LazyArray class extending numpy.ndarray for video and audio loading."""

    def __new__(cls, decord_reader, shape: tuple, dtype: np.dtype):
        buffer = mmap.mmap(-1, dtype.itemsize * math.prod(shape), access=mmap.ACCESS_READ)
        obj = super().__new__(cls, shape, dtype=dtype, buffer=buffer)

        obj.decord_reader = decord_reader
        obj.start_idx = 0
        return obj

    def __getitem__(self, index):
        if isinstance(index, slice):
            indices = list(range(index.start, index.stop))
            return self.decord_reader.get_batch(indices).asnumpy()
        else:
            return np.squeeze(self.decord_reader.get_batch([index]).asnumpy())


class _VideoFileHandler(IHandler):
    """Class for handling the loading and saving of video data."""

    def _get_video_meta(self, fp) -> dict:
        """
        Get video metadata using ffprobe.

        Args:
            fp (Path): The file path of the video.

        Returns:
            dict: A dictionary containing video metadata.
        """
        ffprobe_cmd = [
            "ffprobe",
            "-v",
            "error",
            "-select_streams",
            "v:0",
            "-print_format",
            "json",
            "-show_streams",
            str(fp.resolve()),
        ]
        result = subprocess.run(ffprobe_cmd, capture_output=True, text=True)
        metadata = json.loads(result.stdout)
        return metadata

    def load(self, fp: Path, header_only: bool = False) -> Data:
        """
        Load video data from a file.

        Args:
            fp (Path): The file path of the video.
            header_only (bool): If true only the stream header will be loaded.


        Returns:
            Data: The loaded video data.
        """

        # meta information
        metadata = self._get_video_meta(fp)
        metadata = metadata["streams"][0]
        _width = metadata.get("width")
        _height = metadata.get("height")
        _sample_rate = metadata.get("avg_frame_rate")

        sample_shape = (1, _height, _width, 3)
        duration = int(float(metadata.get("duration")) * 1000)
        sample_rate = eval(_sample_rate) if _sample_rate is not None else None
        num_samples = int(metadata.get("nb_frames"))
        dtype = np.dtype(np.uint8)

        # file loading
        video_reader = decord.VideoReader(str(fp.resolve()), ctx=cpu(0))

        lazy_video_data = None
        if not header_only:
            lazy_video_data = _LazyArray(
                video_reader,
                shape=(num_samples,) + sample_shape[1:],
                dtype=dtype,
            )

        video_ = Video(
            data=lazy_video_data,
            name=fp.stem,
            ext=fp.suffix,
            duration=duration,
            sample_shape=sample_shape,
            num_samples=num_samples,
            sample_rate=sample_rate,
            dtype=dtype,
        )
        return video_

    def save(self, data: Video, fp: Path):
        """
        Save video data to a file.

        Args:
            data (Video): The video data to be saved.
            fp (Path): The file path for saving the data.
        """
        meta_data: StreamMetaData = data.meta_data
        #sample_rate = int(meta_data.sample_rate)
        sample_rate = meta_data.sample_rate
        file_path = str(fp.resolve())

        # TODO: when directly using data.data as a numpy array it is always filled with zeros.
        #  this is an issue with the current implementation of the numpy array buffer in _LazyArray()
        #  creating a list from a buffered array and the a numpy array again solves this problem, but is probably very slow and memory intensive.
        ffmpegio.video.write(
            file_path, sample_rate, np.asarray(list(data.data)), overwrite=True
        )


# AUDIO
class _AudioFileHandler(IHandler):
    """Class for handling the loading and saving of audio data."""

    def _get_audio_meta(self, fp: Path) -> dict:
        """
        Get audio metadata using ffprobe.

        Args:
            fp (Path): The file path of the audio.

        Returns:
            dict: A dictionary containing audio metadata.
        """
        ffprobe_cmd = [
            "ffprobe",
            "-v",
            "error",
            "-print_format",
            "json",
            "-show_streams",
            "-i",
            str(fp.resolve()),
        ]
        result = subprocess.run(ffprobe_cmd, capture_output=True, text=True)
        metadata = json.loads(result.stdout)
        return metadata

    def load(self, fp: Path, header_only: bool = False) -> Data:
        """
        Load audio data from a file.

        Args:
            fp (Path): The file path of the audio.
            header_only (bool): If true only the stream header will be loaded.


        Returns:
            Data: The loaded audio data.
        """
        # meta information
        stream_meta_data = self._get_audio_meta(fp)

        metadata = stream_meta_data["streams"][0]
        _channels = metadata.get("channels")
        _sample_rate = int(metadata.get("sample_rate"))
        _duration = int(float(metadata.get("duration")) * 1000)
        _num_samples = round(_duration * _sample_rate)

        sample_shape = (1, None, _channels)
        duration = _duration
        sample_rate = _sample_rate
        dtype = np.dtype(np.float32)
        num_samples = _num_samples

        # file loading
        lazy_audio_data = None
        if not header_only:
            audio_reader = decord.AudioReader(str(fp.resolve()), ctx=cpu(0))
            lazy_audio_data = _LazyArray(
                audio_reader, shape=audio_reader.shape, dtype=dtype
            )

        audio_ = Audio(
            data=lazy_audio_data,
            duration=duration,
            name=fp.stem,
            ext=fp.suffix,
            sample_shape=sample_shape,
            num_samples=num_samples,
            sample_rate=sample_rate,
            dtype=dtype,
        )
        return audio_

    def save(self, data: Audio, fp: Path):
        """
        Save audio data to a file.

        Args:
            data (Audio): The audio data to be saved.
            fp (Path): The file path for saving the data.
        """
        meta_data: StreamMetaData = data.meta_data
        ffmpegio.audio.write(
            str(fp.resolve()),
            int(meta_data.sample_rate),
            np.swapaxes(np.hstack(data.data), 0, -1),
            overwrite=True,
        )


class FileHandler(IHandler):
    """Class for handling different types of data files."""

    def _get_handler_for_file(
        self, fp
    ) -> Union[
        _AnnotationFileHandler,
        _SSIStreamFileHandler,
        _AudioFileHandler,
        _VideoFileHandler,
    ]:
        """
        Get the appropriate handler for a given file.

        Args:
            fp (Path): The file path.

        Returns:
            IHandler: An instance of the appropriate data handler.
        """
        if not self.data_type:
            ext = fp.suffix[1:]
            if ext in ["annotation"]:
                return _AnnotationFileHandler()
            elif ext in ["stream"]:
                return _SSIStreamFileHandler()
            elif ext in ["wav", "mp3"]:
                return _AudioFileHandler()
            elif ext in ["mp4"]:
                return _VideoFileHandler()
            else:
                raise ValueError(f"Unsupported file extension {fp.suffix}")
        else:
            # TODO provide option to load data with unknown extensions by specifying the datatype
            raise NotImplementedError

    def __init__(self, data_type: int = None):
        self.data_type = data_type

    def load(self, fp: Path, header_only: bool = False) -> Data:
        """
        Load data from a file.

        Args:
            fp (Path): The file path.
            header_only (bool): If true only the stream header will be loaded.

        Returns:
            Data: The loaded data.
        """
        handler = self._get_handler_for_file(fp)
        data = handler.load(fp)
        data.meta_data.expand(FileMetaData(fp))
        return data

    def save(self, data: Stream, fp: Path, overwrite=True, *args, **kwargs):
        """
        Save data to a file.

        Args:
            data (Union[Annotation, SSIStream, Video, Audio]): The data to be saved.
            fp (Path): The file path for saving the data.
            overwrite (bool, optional): Whether to overwrite the file if it exists.
            *args: Variable length argument list.
             **kwargs: Arbitrary keyword arguments.

        Raises:
            FileExistsError: If the file already exists and overwrite is not allowed.
        """
        if fp.exists() and not overwrite:
            raise FileExistsError(f"Cannot write {fp} because file already exists")
        handler = self._get_handler_for_file(fp)
        return handler.save(data, fp, *args, **kwargs)


if __name__ == "__main__":
    # Test cases...
    test_annotations = False
    test_streams = True
    base_dir = Path("../../../test_files/")
    fh = FileHandler()

    """TESTCASE FOR ANNOTATIONS"""
    if test_annotations:

        # ascii read
        discrete_anno_ascii = fh.load(base_dir / "discrete_ascii.annotation")
        continuous_anno_ascii = fh.load(base_dir / "continuous_ascii.annotation")
        free_anno_ascii = fh.load(base_dir / "free_ascii.annotation")

        # binary read
        discrete_anno_binary = fh.load(base_dir / "discrete_binary.annotation")
        continuous_anno_binary = fh.load(base_dir / "continuous_binary.annotation")

        # ascii write
        fh.save(discrete_anno_ascii, base_dir / "new_discrete_ascii.annotation")
        fh.save(continuous_anno_ascii, base_dir / "new_continuous_ascii.annotation")
        fh.save(free_anno_ascii, base_dir / "new_free_ascii.annotation")

        # binary write
        fh.save(
            discrete_anno_binary,
            base_dir / "new_discrete_binary.annotation",
            ftype=SSIFileType.BINARY,
        )
        fh.save(
            continuous_anno_binary,
            base_dir / "new_continuous_binary.annotation",
            ftype=SSIFileType.BINARY,
        )

        # verify
        discrete_anno_ascii_new = fh.load(base_dir / "new_discrete_ascii.annotation")
        continuous_anno_ascii_new = fh.load(
            base_dir / "new_continuous_ascii.annotation"
        )
        free_anno_ascii_new = fh.load(base_dir / "new_free_ascii.annotation")

        # binary read
        discrete_anno_binary_new = fh.load(base_dir / "new_discrete_binary.annotation")
        continuous_anno_binary_new = fh.load(
            base_dir / "new_continuous_binary.annotation"
        )

    """TESTCASE FOR STREAMS"""
    if test_streams:

        # ssistream read
        ssistream_ascii = fh.load(base_dir / "ascii.stream")
        ssistream_binary = fh.load(base_dir / "binary.stream")

        # Replace one dimension with random data
        new_data = ssistream_binary.data.copy()
        replacement_dimension = 0
        random_data = np.random.rand(new_data.shape[replacement_dimension])

        # Generate random data
        new_data[:, replacement_dimension] = random_data
        ssistream_binary.data = new_data
        ssistream_ascii.data = new_data

        # ssistream write
        fh.save(ssistream_ascii, base_dir / "new_ascii.stream", SSIFileType.ASCII)
        fh.save(ssistream_binary, base_dir / "new_binary.stream", SSIFileType.BINARY)

        # audio
        audio = fh.load(base_dir / "test_audio.wav")

        fh.save(audio, base_dir / "new_test_audio.wav")

        new_audio = fh.load(base_dir / "new_test_audio.wav")

        np.allclose(audio.data[0:10000], new_audio.data[0:10000])

        # video
        video = fh.load(base_dir / "test_video.mp4")

        fh.save(video, base_dir / "new_test_video.mp4")

        new_video = fh.load(base_dir / "new_test_video.mp4")

        assert new_video.data[0:30].all() == video.data[0:30].all()
        breakpoint()



""" Module for handling MongoDB data operations related to annotations and streams.

Author:
    Dominik Schiller <dominik.schiller@uni-a.de>
Date:
    18.8.2023
"""
import sys

import numpy as np
import warnings
from datetime import datetime
from pymongo import MongoClient
from pymongo.results import InsertOneResult, UpdateResult

from nova_utils.data.handler.file_handler import FileHandler
from bson.objectid import ObjectId
from pathlib import Path
from nova_utils.data.handler.ihandler import IHandler
from nova_utils.data.annotation import (
    Annotation,
    DiscreteAnnotation,
    DiscreteAnnotationScheme,
    ContinuousAnnotation,
    ContinuousAnnotationScheme,
    FreeAnnotation,
    FreeAnnotationScheme,
)
from nova_utils.utils.anno_utils import (
    convert_ssi_to_label_dtype,
    convert_label_to_ssi_dtype,
)
from nova_utils.data.stream import Stream, SSIStream, Video, Audio, StreamMetaData
from nova_utils.utils.type_definitions import SSILabelDType, SchemeType
from nova_utils.data.session import Session

ANNOTATOR_COLLECTION = "Annotators"
SCHEME_COLLECTION = "Schemes"
STREAM_COLLECTION = "Streams"
ROLE_COLLECTION = "Roles"
ANNOTATION_COLLECTION = "Annotations"
SESSION_COLLECTION = "Sessions"
ANNOTATION_DATA_COLLECTION = "AnnotationData"

# METADATA
class MongoMetaData:
    """
    Metadata for MongoDB connection.

    Attributes:
        ip (str, optional): IP address of the MongoDB server.
        port (int, optional): Port number of the MongoDB server.
        user (str, optional): Username for authentication.
        dataset (str, optional): Name of the dataset.
    """

    def __init__(
        self, ip: str = None, port: int = None, user: str = None, dataset: str = None
    ):
        self.ip = ip
        self.port = port
        self.user = user
        self.dataset = dataset


class MongoAnnotationMetaData(MongoMetaData):
    """
    Metadata for MongoDB annotations.

    Attributes:
        is_locked (bool, optional): Indicates if the annotation is locked.
        is_finished (bool, optional): Indicates if the annotation is finished.
        last_update (bool, optional): Timestamp of the last update.
        annotation_document_id (ObjectId, optional): ID of the annotation document.
        data_document_id (ObjectId, optional): ID of the associated data document.
        **kwargs: Arbitrary keyword arguments.
    """

    def __init__(
        self,
        is_locked: bool = None,
        is_finished: bool = None,
        last_update: bool = None,
        annotation_document_id: ObjectId = None,
        data_document_id: ObjectId = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.is_locked = is_locked
        self.is_finished = is_finished
        self.last_update = last_update
        self.annotation_document_id = annotation_document_id
        self.data_document_id = data_document_id


class MongoStreamMetaData(MongoMetaData):
    """
    Metadata for MongoDB streams.

    Attributes:
        name (str, optional): Name of the stream.
        dim_labels (dict, optional): Dimension labels of the stream.
        file_ext (str, optional): File extension of the stream data.
        is_valid (bool, optional): Indicates if the stream data is valid.
        stream_document_id (ObjectId, optional): ID of the stream document.
        db_sample_rate (float, optional): Sample rate of the stream data in the database.
        type (str, optional): Type of the stream data.
         **kwargs: Arbitrary keyword arguments.
    """

    def __init__(
        self,
        name: str = None,
        dim_labels: dict = None,
        file_ext: str = None,
        is_valid: bool = None,
        stream_document_id: ObjectId = None,
        sr: float = None,
        type: str = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.name = name
        self.dim_labels = dim_labels
        self.file_ext = file_ext
        self.is_valid = is_valid
        self.stream_document_id = stream_document_id
        self.db_sample_rate = sr
        self.type = type


# DATA
class MongoHandler:
    """
    Base class for handling MongoDB connections.

     Args:
        db_host (str, optional): IP address of the MongoDB server.
        db_port (int, optional): Port number of the MongoDB server.
        db_user (str, optional): Username for authentication.
        db_password (str, optional): Password for authentication.

    Attributes:
        data_dir (Path, optional): Path to the data directory for stream files
    """

    def __init__(
        self,
        db_host: str = None,
        db_port: int = None,
        db_user: str = None,
        db_password: str = None,
    ):
        self._client = None
        self._ip = None
        self._port = None
        self._user = None
        if db_host and db_port and db_user and db_password:
            self.connect(db_host, db_port, db_user, db_password)

    def connect(
        self, db_host: str = None, db_port: int = None, db_user: str = None, db_password: str = None
    ):
        """
        Connects to the MongoDB server.

        Args:
            db_host (str): IP address of the MongoDB server.
            db_port (int): Port number of the MongoDB server.
            db_user (str): Username for authentication.
            db_password (str): Password for authentication.
        """
        self._client = MongoClient(host=db_host, port=db_port, username=db_user, password=db_password)
        self._ip = db_host
        self._port = db_port
        self._user = db_user

    @property
    def client(self) -> MongoClient:
        """
        Returns the MongoDB client instance.
        """
        return self._client


class SessionHandler(MongoHandler):
    """
    Handler for loading session data from a MongoDB database.

    This class provides methods to load session information from a MongoDB
    collection and returns a Session object.

    Args:
         (Inherited args from MongoHandler)

    Methods:
        load(dataset: str, session: str) -> Session:
            Load session data from the specified dataset and session name.

    Attributes:
        (Inherited attributes from MongoHandler)

    """

    def load(self, dataset: str, session: str) -> Session:
        """
        Load session data from the specified dataset and session name.

        Args:
            dataset (str): The dataset name as specified in the mongo database
            session (str): The session name as specified in the mongo database

        Returns:
            Session: A Session object containing loaded session information.
            If the session does not exist, an empty Session object is returned.
        """
        result = self.client[dataset][SESSION_COLLECTION].find_one({"name": session})
        if not result:
            return Session()

        # get duration of session in milliseconds
        dur_ms = result.get("duration")
        if dur_ms == 0:
            dur_ms = None
        else:
            dur_ms *= 1000

        return Session(
            dataset=dataset,
            name=result["name"],
            location=result["location"],
            language=result["language"],
            date=result["date"],
            duration=dur_ms,
            is_valid=result["isValid"],
        )


class AnnotationHandler(IHandler, MongoHandler):
    """
    Class for handling download of annotation data from Mongo db.
    """

    def _load_annotation(
        self,
        dataset: str,
        session: str,
        annotator: str,
        role: str,
        scheme: str,
        project: dict = None,
    ) -> dict:
        """
        Load annotation data from MongoDB.

        Args:
            dataset (str): Name of the dataset.
            session (str): Name of the session.
            annotator (str): Name of the annotator.
            role (str): Name of the role.
            scheme (str): Name of the annotation scheme.
            project (dict, optional): Projection for MongoDB query to filter attributes. Defaults to None.

        Returns:
            dict: Loaded annotation data.
        """
        pipeline = [
            {
                "$lookup": {
                    "from": SESSION_COLLECTION,
                    "localField": "session_id",
                    "foreignField": "_id",
                    "as": "session",
                }
            },
            {
                "$lookup": {
                    "from": ANNOTATOR_COLLECTION,
                    "localField": "annotator_id",
                    "foreignField": "_id",
                    "as": "annotator",
                }
            },
            {
                "$lookup": {
                    "from": ROLE_COLLECTION,
                    "localField": "role_id",
                    "foreignField": "_id",
                    "as": "role",
                }
            },
            {
                "$lookup": {
                    "from": SCHEME_COLLECTION,
                    "localField": "scheme_id",
                    "foreignField": "_id",
                    "as": "scheme",
                }
            },
            {
                "$match": {
                    "$and": [
                        {"role.name": role},
                        {"session.name": session},
                        {"annotator.name": annotator},
                        {"scheme.name": scheme},
                    ]
                }
            },
            {
                "$lookup": {
                    "from": "AnnotationData",
                    "localField": "data_id",
                    "foreignField": "_id",
                    "as": "data",
                }
            },
        ]

        # append projection
        if project:
            pipeline.append({"$project": project})

        result = list(self.client[dataset][ANNOTATION_COLLECTION].aggregate(pipeline))
        if not result:
            return {}
        return result[0]

    def _load_scheme(self, dataset:str, scheme:str) -> dict:
        result = self.client[dataset][SCHEME_COLLECTION].find_one({"name": scheme})
        if not result:
            return {}
        return result

    def _update_annotation(
        self,
        dataset: str,
        annotation_id: ObjectId,
        annotation_data_id: ObjectId,
        annotation_data: list[dict],
        is_finished: bool,
        is_locked: bool,
    ) -> UpdateResult:
        """
        Updates existing annotation the Mongo database

        Args:
            dataset (str): Name of the dataset.
            annotation_id (ObjectId): ObjectId of the annotation in the database
            annotation_data_id (ObjectId): ObjectId of the corresponding annotation data object in the database
            annotation_data (list[dict]): List of dictionaries containing the annotation data. Each dictionary represents one sample. Keys must match the annotation types.
            is_finished (bool): Whether the annotation has already been fully completed or not
            is_locked (bool): Whether the annotation should be locked and can therefore not be overwritten anymore.

        Returns:
            UpdateResult: The success status of the update operation
        """
        update_query_annotation = {
            "$set": {
                "date": datetime.now(),
                "isFinished": is_finished,
                "isLocked": is_locked,
            }
        }
        update_query_annotation_data = {"$set": {"labels": annotation_data}}
        success = self.client[dataset][ANNOTATION_COLLECTION].update_one(
            {"_id": annotation_id}, update_query_annotation
        )

        if not success.acknowledged:
            return success

        success = self.client[dataset][ANNOTATION_DATA_COLLECTION].update_one(
            {"_id": annotation_data_id}, update_query_annotation_data
        )

        return success

    def _insert_annotation_data(self, dataset: str, data: list) -> InsertOneResult:
        """
        Insert annotation data into the MongoDB database.

        Args:
            dataset (str): Name of the dataset.
            data (list): List of annotation data to be inserted.

        Returns:
            InsertOneResult: The result of the insertion operation.
        """
        annotation_data = {"labels": data}

        success = self.client[dataset][ANNOTATION_DATA_COLLECTION].insert_one(
            annotation_data
        )
        return success

    def _insert_annotation(
        self,
        dataset: str,
        session_id: ObjectId,
        annotator_id: ObjectId,
        scheme_id: ObjectId,
        role_id: ObjectId,
        data: list,
        is_finished: bool,
        is_locked: bool,
    ):
        """
        Insert annotation and associated annotation data into the MongoDB database.

        Args:
            dataset (str): Name of the dataset.
            session_id (ObjectId): ID of the associated session.
            annotator_id (ObjectId): ID of the annotator.
            scheme_id (ObjectId): ID of the annotation scheme.
            role_id (ObjectId): ID of the role.
            data (list): List of annotation data to be inserted.
            is_finished (bool): Indicates if the annotation is finished.
            is_locked (bool): Indicates if the annotation is locked.

        Returns:
            InsertOneResult: The result of the insertion operation.
        """
        # insert annotation data first
        success = self._insert_annotation_data(dataset, data)
        if not success.acknowledged:
            return success
        else:
            data_id = success.inserted_id

        # insert annotation object
        annotation_document = {
            "session_id": session_id,
            "annotator_id": annotator_id,
            "scheme_id": scheme_id,
            "role_id": role_id,
            "date": datetime.now(),
            "isFinished": is_finished,
            "isLocked": is_locked,
            "data_id": data_id,
        }
        success = self.client[dataset][ANNOTATION_COLLECTION].insert_one(
            annotation_document
        )

        # if the annotation could not be created we delete the annotation data as well
        if not success.acknowledged:
            success = self.client[dataset][ANNOTATION_DATA_COLLECTION].delete_one(
                {"_id": data_id}
            )

        return success


    def load(
        self, dataset: str, scheme: str, session: str, annotator: str, role: str, header_only: bool = False
    ) -> Annotation:
        """
        Load annotation data from MongoDB and create an Annotation object.

        Args:
            dataset (str): Name of the dataset.
            scheme (str): Name of the annotation scheme.
            session (str): Name of the session.
            annotator (str): Name of the annotator.
            role (str): Name of the role.
            header_only (bool): If true only the annotation header will be loaded.


        Returns:
            Annotation: An Annotation object loaded from the database.

        Raises:
            FileNotFoundError: If the requested annotation data is not found in the database.
            TypeError: If the scheme type is unknown.
        """
        # load annotation from mongo db
        if header_only:
            scheme_doc = self._load_scheme(dataset, scheme)

            if not scheme_doc:
                raise FileNotFoundError(
                    f"Scheme not found dataset: {dataset} scheme: {scheme}"
                )

        else:
            anno_doc = self._load_annotation(dataset, session, annotator, role, scheme)

            if not anno_doc:
                raise FileNotFoundError(
                    f"Annotation not found dataset: {dataset} session: {session} annotator: {annotator} role: {role} scheme: {scheme}"
                )

            (anno_data_doc,) = anno_doc["data"]

            # build annotation object
            (scheme_doc,) = anno_doc["scheme"]

        scheme_type = scheme_doc["type"]
        anno_data = None
        anno_duration = 0

        # discrete scheme
        if scheme_type == SchemeType.DISCRETE.name:
            scheme_classes = {l["id"]: l["name"] for l in scheme_doc["labels"]}

            if not header_only:
                anno_data = np.array(
                    [
                        (x["from"], x["to"], x["id"], x["conf"])
                        for x in anno_data_doc["labels"]
                    ],
                    dtype=SSILabelDType.DISCRETE.value,
                )
                anno_data = convert_ssi_to_label_dtype(anno_data, SchemeType.DISCRETE)
                anno_duration = anno_data[-1]["to"] if anno_data.size != 0 else 0

            anno_scheme = DiscreteAnnotationScheme(name=scheme, classes=scheme_classes)
            annotation = DiscreteAnnotation(
                role=role,
                session=session,
                dataset=dataset,
                data=anno_data,
                scheme=anno_scheme,
                annotator=annotator,
                duration=anno_duration,
            )

        # continuous scheme
        elif scheme_type == SchemeType.CONTINUOUS.name:
            sr = scheme_doc["sr"]
            min_val = scheme_doc["min"]
            max_val = scheme_doc["max"]

            if not header_only:
                anno_data = np.array(
                    [(x["score"], x["conf"]) for x in anno_data_doc["labels"]],
                    dtype=SSILabelDType.CONTINUOUS.value,
                )
                anno_data = convert_ssi_to_label_dtype(anno_data, SchemeType.CONTINUOUS)
                anno_duration = len(anno_data_doc["labels"]) / sr

            anno_scheme = ContinuousAnnotationScheme(
                name=scheme, sample_rate=sr, min_val=min_val, max_val=max_val
            )
            annotation = ContinuousAnnotation(
                role=role,
                session=session,
                dataset=dataset,
                data=anno_data,
                scheme=anno_scheme,
                annotator=annotator,
                duration=anno_duration,
            )

            # free scheme

        # free scheme
        elif scheme_type == SchemeType.FREE.name:

            if not header_only:
                anno_data = np.array(
                    [
                        (x["from"], x["to"], x["name"], x["conf"])
                        for x in anno_data_doc["labels"]
                    ],
                    dtype=SSILabelDType.FREE.value,
                )

                anno_data = convert_ssi_to_label_dtype(anno_data, SchemeType.FREE)
                anno_duration = anno_data[-1]["to"] if anno_data.size != 0 else 0

            anno_scheme = FreeAnnotationScheme(name=scheme)
            annotation = FreeAnnotation(
                role=role,
                session=session,
                dataset=dataset,
                data=anno_data,
                scheme=anno_scheme,
                annotator=annotator,
                duration=anno_duration,
            )
        else:
            raise TypeError(f"Unknown scheme type {scheme_type}")

        # setting meta data
        if header_only:
            handler_meta_data = MongoAnnotationMetaData(
                ip=self._ip,
                port=self._port,
                user=self._user,
                dataset=dataset
            )
        else:
            handler_meta_data = MongoAnnotationMetaData(
                ip=self._ip,
                port=self._port,
                user=self._user,
                dataset=dataset,
                is_locked=anno_doc.get("isLocked"),
                is_finished=anno_doc.get("isFinished"),
                annotation_document_id=anno_doc.get("_id"),
                data_document_id=anno_doc.get("data_id"),
                last_update=anno_doc.get("date"),
            )
        annotation.meta_data.expand(handler_meta_data)

        return annotation

    def save(
        self,
        annotation: Annotation,
        dataset: str = None,
        session: str = None,
        annotator: str = None,
        role: str = None,
        is_finished: bool = False,
        is_locked: bool = False,
        overwrite: bool = False,
    ):
        """
        Save an Annotation object to the MongoDB database.

        Args:
            annotation (Annotation): The Annotation object to be saved.
            dataset (str, optional): Name of the dataset. Overwrites the respective attribute from annotation.meta_data if set. Defaults to None.
            session (str, optional): Name of the session. Overwrites the respective attribute from annotation.meta_data if set. Defaults to None.
            annotator (str, optional): Name of the annotator. Overwrites the respective attribute from annotation.meta_data if set. Defaults to None.
            role (str, optional): Name of the role. Overwrites the respective attribute from annotation.meta_data if set. Defaults to None.
            is_finished (bool, optional): Indicates if the annotation is finished. Defaults to False.
            is_locked (bool, optional): Indicates if the annotation is locked. Defaults to False.
            overwrite (bool, optional): If True, overwrite an existing annotation. Defaults to False.

        Returns:
            UpdateResult: The result of the update operation.

        Raises:
            FileExistError: If annotation exists and is locked or annotation exists and overwrite is set to false
        """
        # overwrite default values
        dataset = dataset if not dataset is None else annotation.meta_data.dataset
        session = session if not session is None else annotation.meta_data.session
        annotator = annotator if not annotator is None else annotation.meta_data.annotator
        role = role if not role is None else annotation.meta_data.role
        scheme = annotation.annotation_scheme.name

        anno_data = convert_label_to_ssi_dtype(
            annotation.data, annotation.annotation_scheme.scheme_type
        )

        # TODO check for none values
        anno_data = [
            dict(zip(annotation.annotation_scheme.label_dtype.names, ad.item()))
            for ad in anno_data
        ]

        # load annotation to check if an annotation for the provided criteria already exists in the database
        anno_doc = self._load_annotation(
            dataset,
            session,
            annotator,
            role,
            scheme,
            project={"_id": 1, "isLocked": 1, "data_id": 1},
        )

        # update existing annotations
        if anno_doc:
            if anno_doc["isLocked"]:
                raise FileExistsError(
                    f"Can't overwrite locked annotation dataset: {dataset} session: {session} annotator: {annotator} role: {role} scheme: {scheme}. Because annotation is locked."
                )
            elif not overwrite:
                raise FileExistsError(
                    f"Can't overwrite annotation dataset: {dataset} session: {session} annotator: {annotator} role: {role} scheme: {scheme}. Because overwrite is disabled."
                )
            else:
                warnings.warn(
                    f"Overwriting existing annotation dataset: {dataset} session: {session} annotator: {annotator} role: {role} scheme: {scheme}"
                )

                success = self._update_annotation(
                    dataset=dataset,
                    annotation_id=anno_doc["_id"],
                    annotation_data_id=anno_doc["data_id"],
                    annotation_data=anno_data,
                    is_finished=is_finished,
                    is_locked=is_locked,
                )

        # add new annotation
        else:
            scheme_id = self.client[dataset][SCHEME_COLLECTION].find_one(
                {"name": scheme}
            )["_id"]
            session_id = self.client[dataset][SESSION_COLLECTION].find_one(
                {"name": session}
            )["_id"]
            role_id = self.client[dataset][ROLE_COLLECTION].find_one({"name": role})[
                "_id"
            ]
            annotator_id = self.client[dataset][ANNOTATOR_COLLECTION].find_one(
                {"name": annotator}
            )["_id"]
            success = self._insert_annotation(
                dataset=dataset,
                scheme_id=scheme_id,
                session_id=session_id,
                annotator_id=annotator_id,
                role_id=role_id,
                data=anno_data,
                is_finished=is_finished,
                is_locked=is_locked,
            )

        return success
        # TODO success error handling


class StreamHandler(IHandler, MongoHandler):
    """
    Class for handling download and upload of stream data from MongoDB.
    """

    def __init__(self, *args, data_dir: Path = None, **kwargs):
        """
        Base class for handling MongoDB connections.

         Args:
            ip (str, optional): IP address of the MongoDB server.
            port (int, optional): Port number of the MongoDB server.
            user (str, optional): Username for authentication.
            password (str, optional): Password for authentication.
            data_dir (Path, optional): Path to the data directory for stream files

        Attributes:
            data_dir (Path, optional): Path to the data directory for stream files
            client (MongoClient): The MongoDB client connected to the database. Readonly
        """
        super().__init__(*args, **kwargs)
        self.data_dir = Path(data_dir)

    def _load_stream(
        self,
        dataset: str,
        stream_name: str,
    ) -> dict:
        """
        Load stream data from MongoDB.

        Args:
            dataset (str): Name of the dataset.
            stream_name (str): Name of the stream.

        Returns:
            dict: Loaded stream data.
        """
        result = self.client[dataset][STREAM_COLLECTION].find_one({"name": stream_name})
        if not result:
            return {}
        return result

    def load(self, dataset: str, session: str, role: str, name: str, header_only: bool = False) -> Stream:
        """
        Load a Stream object from MongoDB and create a Stream instance.

        Args:
            dataset (str): Name of the dataset.
            session (str): Name of the session.
            role (str): Name of the role.
            name (str): Name of the stream.
            header_only (bool): If true only the stream header will be loaded.

        Returns:
            Stream: A Stream object loaded from the database.

        Raises:
            ValueError: If the requested stream is not found for the given dataset.
            FileNotFoundError: If the data directory is not set or the file is not found on disc.
        """
        result = self._load_stream(dataset=dataset, stream_name=name)
        if not result:
            raise FileNotFoundError(f"No stream {name} found for dataset {dataset}")
        if not self.data_dir:
            raise FileNotFoundError("Data directory was not set. Can't access files")

        file_path = Path(
            self.data_dir
            / dataset
            / session
            / (role + "." + result["name"] + "." + result["fileExt"])
        )

        if not file_path.is_file():
            raise FileNotFoundError(f"No such file {file_path}")

        # data
        data = FileHandler().load(file_path, header_only = header_only)
        assert isinstance(data, Stream)

        # meta data
        data.meta_data.role = role
        data.meta_data.name = name
        data.meta_data.session = session

        handler_meta_data = MongoStreamMetaData(
            ip=self._ip,
            port=self._port,
            user=self._user,
            name=result.get("name"),
            dim_labels=result.get("dimLabels"),
            file_ext=result.get("fileExt"),
            is_valid=result.get("isValid"),
            stream_document_id=result.get("_id"),
            sr=result.get("sr"),
            type=result.get("type"),
            dataset=dataset
        )
        data.meta_data.expand(handler_meta_data)

        return data

    def save(
        self,
        stream: Stream,
        dataset: str = None,
        session: str = None,
        role: str = None ,
        name: str = None,
        data_type: str = 'feature',
        file_ext: str = None,
        dim_labels: list = None,
        is_valid: bool = True,
    ):
        """
        Save a Stream object to the MongoDB database and store associated file.

        Args:
            stream (Stream): The Stream object to be saved.
            dataset (str): Name of the dataset. Overwrites the respective attribute from stream.meta_data if set. Defaults to None.
            session (str): Name of the session. Overwrites the respective attribute from stream.meta_data if set. Defaults to None.
            role (str): Name of the role. Overwrites the respective attribute from stream.meta_data if set. Defaults to None.
            name (str): Name of the stream. Overwrites the respective attribute from stream.meta_data if set. Defaults to None.
            data_type (str): Media type of the stream data as specified in NOVA-DB.
            file_ext (str, optional): File extension. Defaults to None.
            dim_labels (list[dict], optional): Dimension labels. Defaults to None.
            is_valid (bool, optional): Indicates if the stream data is valid. Defaults to True.

        Raises:
            FileNotFoundError: If the data directory is not set.
        """


        # save file to disk
        dataset = dataset if not dataset is None else stream.meta_data.dataset
        session = session if not session is None else stream.meta_data.session
        role = role if not role is None else stream.meta_data.role
        name = name if not name is None else stream.meta_data.name
        file_ext = file_ext if not file_ext is None else stream.meta_data.ext
        if isinstance(stream, SSIStream):
            dim_labels = dim_labels if not dim_labels is None else stream.meta_data.dim_labels
        else:
            dim_labels = None

        if not self.data_dir:
            raise FileNotFoundError("Data directory was not set. Can't access files")

        if not self.data_dir.is_dir():
            raise NotADirectoryError(f"Specified data directory {self.data_dir} is not a directory.")

        file_name = role + "." + name + file_ext
        file_path = self.data_dir / dataset / session / file_name

        FileHandler().save(stream, file_path)

        meta_data: StreamMetaData = stream.meta_data

        # write db entry
        stream_document = {
            "fileExt": file_ext.strip('.'),
            "name": name,
            "sr": meta_data.sample_rate,
            "type": data_type,
            "dimlabels": dim_labels if dim_labels else [],
            "isValid": is_valid,
        }

        # check if stream exists
        result = self.client[dataset][STREAM_COLLECTION].find_one({"name": name})

        # update existing
        if result:
            update_query_annotation = {"$set": stream_document}
            self.client[dataset][STREAM_COLLECTION].update_one(
                {"_id": result["_id"]}, update_query_annotation
            )

        # insert new
        else:
            self.client[dataset][STREAM_COLLECTION].insert_one(stream_document)


if __name__ == "__main__":
    import os
    import random
    from time import perf_counter
    from dotenv import load_dotenv

    test_annotations = False
    test_streams = True

    load_dotenv("../../../.env")
    IP = os.getenv("NOVA_IP", "")
    PORT = int(os.getenv("NOVA_PORT", 0))
    USER = os.getenv("NOVA_USER", "")
    PASSWORD = os.getenv("NOVA_PASSWORD", "")
    DATA_DIR = os.getenv("NOVA_DATA_DIR", None)

    if test_annotations:
        amh = AnnotationHandler(db_host=IP, db_port=PORT, db_user=USER, db_password=PASSWORD)

        # load
        fs = "Loading {} took {}ms"
        t_start = perf_counter()
        discrete_anno = amh.load(
            dataset="test",
            scheme="diarization",
            annotator="schildom",
            session="04_Oesterreich_test",
            role="testrole2",
            header_only=True
        )
        t_stop = perf_counter()
        print(fs.format("Discrete annotation", int((t_stop - t_start) * 1000)))

        t_start = perf_counter()
        continuous_anno = amh.load(
            dataset="test",
            scheme="arousal",
            annotator="emow2v",
            session="01_AffWild2_video1",
            role="testrole",
        )
        t_stop = perf_counter()
        print(fs.format("Continuous annotation", int((t_stop - t_start) * 1000)))

        t_start = perf_counter()
        free_anno = amh.load(
            dataset="test",
            scheme="transcript",
            annotator="whisperx",
            session="04_Oesterreich_test",
            role="testrole",
        )
        t_stop = perf_counter()
        print(fs.format("Free annotation", int((t_stop - t_start) * 1000)))

        # save
        fs = "Saving {} took {}ms"
        t_start = perf_counter()
        amh.save(
            dataset="test",
            annotation=discrete_anno,
            session="04_Oesterreich_test",
            annotator="testuser",
            role="testrole",
            overwrite=True,
        )
        t_stop = perf_counter()
        print(fs.format("Discrete annotation", int((t_stop - t_start) * 1000)))

        t_start = perf_counter()
        # amh.save(continuous_anno, annotator="testuser", overwrite=True)
        t_stop = perf_counter()
        print(fs.format("Continuous annotation", int((t_stop - t_start) * 1000)))

        t_start = perf_counter()
        # amh.save(free_anno, annotator="testuser", overwrite=True)
        t_stop = perf_counter()
        print(fs.format("Free annotation", int((t_stop - t_start) * 1000)))

    if test_streams:

        smh = StreamHandler(
            db_host=IP, db_port=PORT, db_user=USER, db_password=PASSWORD, data_dir=Path(DATA_DIR)
        )

        # Stream
        fs = "Loading {} took {}ms"
        t_start = perf_counter()
        feature_stream = smh.load(
            dataset="test",
            session="04_Oesterreich_test",
            role="testrole",
            name="arousal.synchrony[testrole]",
            header_only=True
        )
        t_stop = perf_counter()
        print(fs.format("Video", int((t_stop - t_start) * 1000)))

        suffix = "_testing"
        feature_stream.sample_rate = random.uniform(0, 16000)
        smh.save(
            stream=feature_stream,
            dataset="test",
            session="04_Oesterreich_test",
            role="testrole",
            name="arousal.synchrony[testrole]" + suffix,
            data_type="video",
            dim_labels=[{"id": 1, "name": "hallo"}, {"id": 2, "name": "nope"}],
        )

        # Audio
        t_start = perf_counter()
        audio_stream = smh.load(
            dataset="test", session="01_AffWild2_video1", role="testrole", name="audio"
        )
        t_stop = perf_counter()
        print(fs.format("Audio", int((t_stop - t_start) * 1000)))

        # Video
        t_start = perf_counter()
        video_stream = smh.load(
            dataset="test", session="01_AffWild2_video1", role="testrole", name="video"
        )
        t_stop = perf_counter()
        print(fs.format("Video", int((t_stop - t_start) * 1000)))

        breakpoint()



""" Interface class for all data handlers

Author:
    Dominik Schiller <dominik.schiller@uni-a.de>
Date:
    18.8.2023

"""

from nova_utils.data.data import Data
from abc import ABC, abstractmethod


class IHandler(ABC):
    """
    Abstract base class for data handling operations.

    Subclasses of IHandler must implement the methods 'load' and 'save'
    to handle data loading and saving operations, respectively.
    """

    @abstractmethod
    def load(self, *args, **kwargs) -> Data:
        """
        Load data using specified arguments and keyword arguments.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            Data: A Data object representing the loaded data.

        Raises:
            NotImplementedError: If the method is not implemented by a subclass.
        """
        raise NotImplementedError()

    @abstractmethod
    def save(self, *args, **kwargs):
        """
        Save data using specified arguments and keyword arguments.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Raises:
            NotImplementedError: If the method is not implemented by a subclass.
        """
        raise NotImplementedError()



"""Utility module to parse and process strings
Author:
    Dominik Schiller <dominik.schiller@uni-a.de>
Date:
    14.9.2023

"""
from typing import Union
from enum import Enum

def parse_time_string_to_ms(frame: Union[str, int, float, None]) -> int:
    """
    Parse a time string or value to milliseconds.

    This function takes a frame as input, which can be specified as a string in milliseconds (e.g., "100ms"),
    as a string in seconds (e.g., "2s"), as a float in seconds, as an integer in milliseconds, or as None.
    It converts the frame to an integer value in milliseconds.

    Args:
        frame (Union[str, int, float, None]): The frame value to parse.

    Returns:
        int: The frame value in milliseconds. 0 if frame is None.

    Raises:
        ValueError: If the input format for frame is invalid.

    """
    if frame is None:
        return 0

    # if frame is specified milliseconds as string
    if str(frame).endswith("ms"):
        try:
            return int(frame[:-2])
        except ValueError:
            raise ValueError(
                "Invalid input format for frame in milliseconds: {}".format(frame)
            )
    # if frame is specified in seconds as string
    elif str(frame).endswith("s"):
        try:
            frame_s = float(frame[:-1])
            return int(frame_s * 1000)
        except ValueError:
            raise ValueError(
                "Invalid input format for frame in seconds: {}".format(frame)
            )
    # if type is float we assume the input will be seconds
    elif isinstance(frame, float) or "." in str(frame):
        try:
            print(
                "WARNING: Automatically inferred type for frame {} is float.".format(
                    frame
                )
            )
            return int(1000 * float(frame))
        except ValueError:
            raise ValueError("Invalid input format for frame: {}".format(frame))

    # if type is int we assume the input will be milliseconds
    elif isinstance(frame, int) or (isinstance(frame, str) and frame.isdigit()):
        try:
            print(
                "WARNING: Automatically inferred type for frame {} is int.".format(
                    frame
                )
            )
            return int(frame)
        except ValueError:
            raise ValueError("Invalid input format for frame: {}".format(frame))
    else:
        raise ValueError("Invalid input format for frame: {}".format(frame))

def string_to_enum(enum: Enum, string: str):
    """
    Convert a string to an enum value.

    This function takes an enum and a string and returns the corresponding enum value. If the string does not match any enum value,
    a ValueError is raised.

    Args:
        enum (Enum): The enum to search in.
        string (str): The string to convert to an enum value.

    Returns:
        Enum: The enum value corresponding to the input string.

    Raises:
        ValueError: If the input string does not match any enum value in the specified enum.

    """
    for e in enum:
        if e.name == string:
            return e
    raise ValueError('{} not part of enumeration  {}'.format(string, enum))

def string_to_bool(string:str) -> bool:
    '''
    Parses a given input string to a boolean value
    Args:
        string (str): Input string

    Returns:
        bool: The boolean value of the string
    '''
    if string in ['True', 'true', '1']:
        return True
    else:
        return False

def parse_nova_option_string(option_string: str) -> dict:
    """
    Converts a server-module option string to dictionary.

    This function takes an option string as send by nova and converts it to dictionary containing the option name as key and the according value as value.

    Args:
        option_string (str): The option string.

    Returns:
        Enum: The enum value corresponding to the input string.
    """
    options = {}
    print('Parsing options')
    if option_string:
        opts = option_string.split(';')
        for option in opts:
            k,v = option.split('=')
            if v in ("True", "False"):
                options[k] = True if v == "True" else False

            elif v == "None":
                options[k] = None
            else:
                try:
                    options[k] = int(v)
                except:
                    try:
                        options[k] = float(v)
                    except:
                         options[k] = v
            print('\t' + k + "=" + v)
    print('...done')
    return options


from abc import ABC, abstractmethod

import numpy as np

from nova_utils.data.annotation import Annotation
from nova_utils.data.stream import Stream
from nova_utils.utils.ssi_xml_utils import ModelIO, Trainer
from nova_utils.utils.string_utils import parse_nova_option_string


class Processor(ABC):
    """
    Base class of a data processor. This interface builds the foundation for all data processing classes.
    """

    # List of dependencies that need to be installed when the script is loaded
    DEPENDENCIES = []

    # Flag to indicate whether the processed input belongs to one role or to multiple roles
    SINGLE_ROLE_INPUT = True

    # Prefix for provided datastreams that are missing the id-tag
    UNKNOWN_ID = '<unk>_'

    # TODO read trainer or chain file for default options
    def __init__(self, model_io: list[ModelIO], opts: dict, trainer: Trainer = None):
        self.model = None
        self.data = None
        self.output = None
        self.options = opts
        self.model_io = model_io
        self.trainer = trainer

    def preprocess_sample(self, sample: dict):
        """Preprocess data to convert between nova-server dataset iterator item to the raw model input as required in process_sample.

        Args:
            sample :
        """
        return list(sample.values())[0]

    def process_sample(self, sample):
        """Applying processing steps (e.g. feature extraction, data prediction etc... ) to the provided data."""
        return sample

    def postprocess_sample(self, sample):
        """Apply any optional postprocessing to the data (e.g. scaling, mapping etc...)"""
        return sample

    def process_data(self, ds_iter) -> dict:
        """Returning a dictionary that contains the original keys from the dataset iterator and a list of processed samples as value. Can be overwritten to customize the processing"""
        self.ds_iter = ds_iter

        # Get all data streams of type "input" that match an id from the modules trainer file
        processed = {
            k: []
            for k in [
                d.get("id")
                for d in ds_iter.data
                if d.get("type") == "input" and d.get("id") in [mio.io_id for mio in self.model_io]
            ]
        }

        for sample in ds_iter:
            for id, output_list in processed.items():
                data_for_id = {id: sample[id]}
                out = self.preprocess_sample(data_for_id)
                out = self.process_sample(out)
                out = self.postprocess_sample(out)
                output_list.append(out)

        return processed


class Trainer(Processor):
    """
    Base class of a Trainer. Implement this interface in your own class to build a model that is trainable from within nova
    """

    """Includes all the necessary files to run this script"""

    @abstractmethod
    def train(self):
        """Trains a model with the given data."""
        raise NotImplemented

    @abstractmethod
    def save(self, path) -> str:
        """Stores the weights of the given model at the given path. Returns the path of the weights."""
        raise NotImplemented

    @abstractmethod
    def load(self, path):
        """Loads a model with the given path. Returns this model."""
        raise NotImplemented


class Predictor(Processor):
    """
    Base class of a data predictor. Implement this interface if you want to write annotations to a database
    """

    @abstractmethod
    def to_anno(self, data) -> list[Annotation]:
        """Converts the output of process_data to the correct annotation format to upload them to the database.
        !THE OUTPUT FORMAT OF THIS FUNCTION IS NOT YET FULLY DEFINED AND WILL CHANGE IN FUTURE RELEASES!

        Args:
            data (object): Data output of process_data function

        Returns:
            list: A list of annotation objects
        """
        raise NotImplemented


class Extractor(Processor):
    """
    Base class of a feature extractor. Implement this interface in your own class to build a feature extractor.
    """

    @property
    @abstractmethod
    def chainable(self):
        """Whether this extraction module can be followed by other extractors. If set to True 'to_ds_iterable()' must be implemented"""
        return False

    @abstractmethod
    def to_stream(self, data: object) -> list[Stream]:
        """Converts the return value from process_data() to data stream chunk that can be processed by nova-server.

        Args:
            data (object): The data as returned by the process_data function of the Processor class


        Returns:
            list: A list of stream objects.

        """
        raise NotImplemented



"""Standalone script for data extraction

Author:
    Dominik Schiller <dominik.schiller@uni-a.de>
Date:
    06.09.2023

This script utilizes novaserver modules to extract datastreams from either NOVA-DB annotations and datastreams.

.. argparse::
   :module: nova_utils.scripts.predict
   :func: parser
   :prog: nu-predict

Returns:
    None

Example:
    >>> nu-extract --dataset "test" --db_host "127.0.0.1" --db_port "37317" --db_user "my_user" --db_password "my_password" --chain_file_path "test\\test_extract.chain" --sessions "[\"test_session_1\", \"test_session_2\"]" --data "[{\"src\": \"db:anno\", \"scheme\": \"transcript\", \"annotator\": \"test\", \"role\": \"testrole\"}]" --frame_size "0" --left_context "0" --right_context "0" --job_i_d "test_job" --opt_str "num_speakers=2;speaker_ids=testrole,testrole2" --cml_dir "./cml" --data_dir "./data" --log_dir "./log" --cache_dir "./cache" --tmp_dir "./tmp"
"""

import argparse
from pathlib import Path, PureWindowsPath
from nova_utils.utils.ssi_xml_utils import Chain, ChainLink
from nova_utils.data.provider.nova_iterator import NovaIterator
from nova_utils.scripts.parsers import (
    nova_db_parser,
    nova_iterator_parser,
    nova_server_module_parser,
)
from nova_utils.data.handler.mongo_handler import StreamHandler
from importlib.machinery import SourceFileLoader

# Main parser for predict specific options
parser = argparse.ArgumentParser(
    description="Use a provided nova-server module for inference and save results to NOVA-DB",
    parents=[nova_db_parser, nova_iterator_parser, nova_server_module_parser],
)
parser.add_argument(
    "--chain_file_path",
    type=str,
    required=True,
    help="Path to the chain file using Windows UNC-Style",
)

def _main():

    args, _ = parser.parse_known_args()

    # Create argument groups
    db_args, _ = nova_db_parser.parse_known_args()
    iter_args, _ = nova_iterator_parser.parse_known_args()
    module_args, _ = nova_server_module_parser.parse_known_args()

    caught_ex = False

    # Load chain
    chain = Chain()
    chain_file_path = Path(module_args.cml_dir).joinpath(
        PureWindowsPath(args.chain_file_path)
    )
    if not chain_file_path.is_file():
        raise FileNotFoundError(f"Chain file not available: {chain_file_path}")
    else:
        chain.load_from_file(chain_file_path)
        print("Chain successfully loaded.")


    # Build data loaders
    sessions = iter_args.sessions
    iterators = []

    # TODO split for role if multirole input is false
    args = {**vars(db_args), **vars(iter_args)}

    for session in sessions:
        print(session)
        args["sessions"] = [session]
        ni = NovaIterator(**args)
        iterators.append(ni)
    print("Data iterators initialized")

    # Init database handler
    stream_handler = StreamHandler(**vars(db_args), data_dir=iter_args.data_dir)

    # Iterate over all sessions
    for ds_iter in iterators:
        print(f"Extract data for session {ds_iter.sessions[0]}...")

        # Iterate over all chain links
        cl: ChainLink
        for i, cl in enumerate(chain.links):
            model_script_path = chain_file_path.parent / cl.script
            source = SourceFileLoader(
                "ns_cl_" + model_script_path.stem, str(model_script_path)
            ).load_module()
            extractor_class = getattr(source, cl.create)
            extractor = extractor_class()
            print(f"Model {cl.create} created")

            # Check if there will be more chain links to execute
            if i + 1 != len(chain.links):
                raise NotImplementedError('Chains with more than one element will be supported in future releases.')
                # Assert chainability
                # if not extractor.chainable:
                #     raise AssertionError(
                #         "Extraction module does not not support further chaining but is not the last link in chain."
                #     )
                # # TODO implement nova data types in the server module
                # # TODO process data and write tmp session files to disk if necessary
                # print("Extract data...")
                #
                # try:
                #     data = extractor.process_data(ds_iter)
                #     ds_iter = extractor.to_ds_iterable(data)
                # except Exception as e:
                #     print(
                #         f"\tIterator exited with error: '{str(e)}'. Continuing with next session."
                #     )
                #     caught_ex = True
                #     continue
                #
                # print("...done")

            # Last element of chain
            else:
                try:
                    data = extractor.process_data(ds_iter)
                    streams = extractor.to_stream(data)
                except FileNotFoundError as e:
                    print(
                        f"\tIterator exited with error: '{str(e)}'. Continuing with next session."
                    )
                    caught_ex = True
                    break

                print("...done")

                print("Saving streams to disk...")
                try:
                    for stream in streams:
                        print(f"\t{stream.meta_data.name}")
                        stream_handler.save(stream)
                except FileExistsError as e:
                    print(
                        f"\tCould note save stream: '{str(e)}'. Continuing with next session."
                    )
                    caught_ex = True
                    continue
                finally:
                    print("...done")

    print("Prediction completed!")
    if caught_ex:
        print(
            "Prediction job encountered errors for some sessions. Check logs for details."
        )
        exit(1)


if __name__ == "__main__":
    _main()
 


"""Standalone script for data predictions

Author:
    Dominik Schiller <dominik.schiller@uni-a.de>
Date:
    06.09.2023

This script performs predictions for annotations in the NOVA database using a provided nova-server module for inference and saves the results to the NOVA-DB.

.. argparse::
   :module: nova_utils.scripts.predict
   :func: parser
   :prog: nu-predict

Returns:
    None

Example:
    >>> nu-predict --dataset "test" --db_host "127.0.0.1" --db_port "37317" --db_user "my_user" --db_password "my_password" --trainer_file_path "test\\test_predict.trainer" --sessions "[\"test_session_1\", \"test_session_2\"]" --data "[{\"src\": \"db:anno\", \"scheme\": \"transcript\", \"annotator\": \"test\", \"role\": \"testrole\"}]" --frame_size "0" --left_context "0" --right_context "0" --job_i_d "test_job" --opt_str "num_speakers=2;speaker_ids=testrole,testrole2" --cml_dir "./cml" --data_dir "./data" --log_dir "./log" --cache_dir "./cache" --tmp_dir "./tmp"
"""

import argparse
from pathlib import Path, PureWindowsPath
from nova_utils.utils import ssi_xml_utils
from nova_utils.data.provider.nova_iterator import NovaIterator
from nova_utils.scripts.parsers import (
    nova_db_parser,
    nova_iterator_parser,
    nova_server_module_parser,
)
from nova_utils.data.handler.mongo_handler import AnnotationHandler
from importlib.machinery import SourceFileLoader

# Main parser for predict specific options
parser = argparse.ArgumentParser(
    description="Use a provided nova-server module for inference and save results to NOVA-DB",
    parents=[nova_db_parser, nova_iterator_parser, nova_server_module_parser],
)
parser.add_argument(
    "--trainer_file_path",
    type=str,
    required=True,
    help="Path to the trainer file using Windows UNC-Style",
)

def _main():

    args, _ = parser.parse_known_args()

    # Create argument groups
    db_args, _ = nova_db_parser.parse_known_args()
    iter_args, _ = nova_iterator_parser.parse_known_args()
    module_args, _ = nova_server_module_parser.parse_known_args()

    caught_ex = False

    # Load trainer
    trainer = ssi_xml_utils.Trainer()
    trainer_file_path = Path(module_args.cml_dir).joinpath(
        PureWindowsPath(args.trainer_file_path)
    )
    if not trainer_file_path.is_file():
        raise FileNotFoundError(f"Trainer file not available: {trainer_file_path}")
    else:
        trainer.load_from_file(trainer_file_path)
        print("Trainer successfully loaded.")

    # Load module
    if not trainer.model_script_path:
        raise ValueError('Trainer has no attribute "script" in model tag.')

    model_script_path = (
            trainer_file_path.parent / PureWindowsPath(trainer.model_script_path)
    ).resolve()
    source = SourceFileLoader(
        "ns_cl_" + model_script_path.stem, str(model_script_path)
    ).load_module()
    print(f"Trainer module {Path(model_script_path).name} loaded")

    trainer_class = getattr(source, trainer.model_create)
    predictor = trainer_class()
    print(f"Model {trainer.model_create} created")

    # TODO implement SSI interface
    # If the module implements the Trainer interface load weights
    # if isinstance(predictor, iTrainer):
    #     # Load Model
    #     model_weight_path = (
    #             trainer_file_path.parent / trainer.model_weights_path
    #     )
    #     logger.info(f"Loading weights from {model_weight_path}")
    #     predictor.load(model_weight_path)
    #     logger.info("Model loaded.")

    # TODO split for role if multirole input is false
    # Build data loaders
    sessions = iter_args.sessions
    iterators = []
    args = {**vars(db_args), **vars(iter_args)}

    for session in sessions:
        print(session)
        args["sessions"] = [session]
        ni = NovaIterator(**args)
        iterators.append(ni)
    print("Data iterators initialized")

    # Init database handler
    db_handler = AnnotationHandler(**vars(db_args))

    # Iterate over all sessions
    for ds_iter in iterators:
        print(f"Predict session {ds_iter.sessions[0]}...")
        try:
            data = predictor.process_data(ds_iter)
            annos = predictor.to_anno(data)
        except FileNotFoundError as e:
            print(
                f"\tIterator exited with error: '{str(e)}'. Continuing with next session."
            )
            caught_ex = True
            continue
        finally:
            print("...done")

        print("Saving predictions to database...")

        try:
            for anno in annos:
                db_handler.save(anno)
        except FileExistsError as e:
            print(
                f"\tCould note save annotation: '{str(e)}'. Continuing with next session."
            )
            caught_ex = True
            continue
        finally:
            print("...done")

    print("Prediction completed!")
    if caught_ex:
        print(
            "Prediction job encountered errors for some sessions. Check logs for details."
        )
        exit(1)


if __name__ == "__main__":
    _main()



"""nova_utils.data.handler

Author:
    Dominik Schiller <dominik.schiller@uni-a.de>
Date:
    06th September 2023

The nova-utils data handler package implements different handler to load and save data to various backends.
"""


"""Data Storage Class for session specific data
Author: Dominik Schiller <dominik.schiller@uni-a.de>
Date: 22.8.2023
"""
import datetime

class Session:
    """
    Class to stores all information belonging to a specific session during processing

    Attributes:
        data (dict, optional):  Annotation or Stream data that can be processed by a module.
        dataset (str, optional): The dataset or category the session belongs to.
        name (str, optional): The name or title of the session.
        duration (int, optional): The duration of the session in minutes.
        location (str, optional): The location or venue of the session.
        language (str, optional): The language used in the session.
        date (datetime, optional): The date and time of the session.
        is_valid (bool, optional): Whether the session is considered valid.
        extra_data (dict, optional): Additional stream or annotation information for the session. Might only contain meta information.
        output_data_templates (dict, optional): Empty data objects that contain meta information for outputs. Modules can fill the data here.

    Args:
        input_data (dict, optional): Annotation or Stream data that can be processed by a module.
        dataset (str, optional): The dataset or category the session belongs to.
        name (str, optional): The name or title of the session.
        duration (int, optional): The duration of the session in milliseconds.
        location (str, optional): The location or venue of the session.
        language (str, optional): The language used in the session.
        date (datetime, optional): The date and time of the session.
        is_valid (bool, optional): Whether the session is considered valid.
        extra_data (dict, optional): Additional stream or annotation information for the session. Might only contain meta information.
        output_data_templates (dict, optional): Empty data objects that contain meta information for outputs. Modules can fill the data here.
    """
    def __init__(self, input_data: dict = None, dataset: str = None, name: str = None, duration: int = None, location: str = None, language: str = None, date: datetime = None, is_valid: bool = True, extra_data: dict = None, output_data_templates: dict = None):
        self.input_data = input_data
        self.dataset = dataset
        self.name = name
        self.duration = duration
        self.location = location
        self.language = language
        self.date = date
        self.is_valid = is_valid
        self.extra_data = extra_data
        self.output_data_templates = output_data_templates


"""Definition of all Stream classes and Metadata
Author: Dominik Schiller <dominik.schiller@uni-a.de>
Date: 18.8.2023
"""

import numpy as np
from typing import Union
from nova_utils.data.data import DynamicData
from nova_utils.utils.type_definitions import SSINPDataType
from nova_utils.utils.stream_utils import time_to_sample_interval


class StreamMetaData:
    """
    Metadata for a data stream, providing information about the stream properties.

    Attributes:
        name (str): Name of the stream.
        ext (str): File extension of the stream including the leading '.'
        duration (float): Duration of the stream in seconds.
        sample_shape (tuple): Shape of individual samples in the stream.
        num_samples (int): Total number of samples in the stream.
        sample_rate (float): Sampling rate of the stream in Hz.
        dtype (np.dtype): Data type of the samples.

    Args:
        name (str): Name of the stream.
        ext (str): File extension of the stream including the leading '.'
        duration (float, optional): Duration of the stream in seconds.
        sample_shape (tuple, optional): Shape of individual samples in the stream.
        num_samples (int, optional): Number of samples in the stream.
        sample_rate (float, optional): Sampling rate of the stream.
        dtype (np.dtype, optional): Data type of the samples.
    """

    def __init__(
        self,
        name: str = None,
        ext: str = None,
        duration: float = None,
        sample_shape: tuple = None,
        num_samples: int = None,
        sample_rate: float = None,
        dtype: np.dtype = None,

    ):
        """
        Initialize a StreamMetaData instance with stream properties.
        """
        self.name = name
        self.ext = ext
        self.duration = duration
        self.sample_shape = sample_shape
        self.num_samples = num_samples
        self.sample_rate = sample_rate
        self.dtype = dtype


class SSIStreamMetaData:
    """
    Metadata specific to SSI stream files.

    Attributes:
        chunks (np.ndarray): Chunks of the SSI stream with 'from', 'to', 'byte', and 'num' properties.
        dimlabels (list[dict]): List of dictionaries mapping an integer id as key (stream index) to a descriptive string as value.

    Args:
        chunks (np.ndarray): Chunks of the SSI stream with 'from', 'to', 'byte', and 'num' properties.
        dimlabels (list[dict]): List of dictionaries mapping an integer id as key (stream index) to a descriptive string as value.

    """

    CHUNK_DTYPE = np.dtype(
        [
            ("from", SSINPDataType.FLOAT.value),
            ("to", SSINPDataType.FLOAT.value),
            ("byte", SSINPDataType.INT.value),
            ("num", SSINPDataType.INT.value),
        ]
    )

    def __init__(self, chunks: np.ndarray, dim_labels: list = None):
        """
        Initialize an SSIStreamMetaData instance with chunks information.
        """
        self.chunks = chunks
        self.dim_labels = dim_labels


class Stream(DynamicData):
    """
    A class representing a generic data stream along with associated metadata.

    This class extends the DynamicData class and implements methods for working
    with stream data.

    Attributes:
        (Inherits attributes from DynamicData.)

    Args:
        data (np.ndarray): The data stream.
        sample_rate (float): Sampling rate of the stream.
        duration (float, optional): Duration of the stream in seconds.Will be added to metadata.
        sample_shape (tuple, optional): Shape of individual samples in the stream. Will be added to metadata.
        num_samples (int, optional): Number of samples in the stream. Will be added to metadata.
        dtype (np.dtype, optional): Data type of the samples. Will be added to metadata. Defaults to np.float32 .
        name (str, optional): Name of the stream.
        ext (str, optional): File extension of the stream including the leading '.'. Defaults to '.stream', '.mp4' or '.wav' depending on the Streamtype.
        **kwargs: Additional keyword arguments for DynamicData.

    """

    def __init__(
        self,
        data: Union[np.ndarray, None],
        sample_rate: float,
        name: str = None,
        ext: str = None,
        duration: float = None,
        sample_shape: tuple = None,
        num_samples: int = None,
        dtype: np.dtype = SSINPDataType.FLOAT.value,
        **kwargs
    ):
        """
        Initialize a Stream instance with stream data and metadata.
        """
        super().__init__(data=data, **kwargs)

        # Add Metadata
        if ext is None:
            if isinstance(self, SSIStream):
                ext = '.stream'
            if isinstance(self, Video):
                ext = '.mp4'
            if isinstance(self, Audio):
                ext = '.wav'

        stream_meta_data = StreamMetaData(
            name, ext, duration, sample_shape, num_samples, sample_rate, dtype
        )
        self.meta_data.expand(stream_meta_data)

    def sample_from_interval(self, start: int, end: int) -> np.ndarray:
        """
        Abstract method to sample data from within the specified interval.

        Args:
            start (int): The start index of the interval.
            end (int): The end index of the interval.

        Returns:
            np.ndarray: The sampled data within the interval.
        """

        start_sample, end_sample = time_to_sample_interval(start, end, self.meta_data.sample_rate)
        return self.data[start_sample : end_sample]


class SSIStream(Stream):
    """
    A class representing an SSI data stream.

    This class extends the Stream class with additional attributes specific to SSI streams.

    Attributes:
        (Inherits attributes from Stream.)
        CHUNK_DTYPE (np.dtype): Data type definition for SSI stream chunks.

    Args:
        data (np.ndarray): The SSI stream data. Shape is (num_samples,) + (sample_shape,)
        sample_rate (float): Sampling rate of the SSI stream.
        chunks (np.ndarray, optional): Chunks of the SSI stream.
        **kwargs: Additional keyword arguments for Stream.

    Methods:
        (No additional methods specified in the provided code.)
    """

    CHUNK_DTYPE = np.dtype(
        [
            ("from", SSINPDataType.FLOAT.value), # start of chunk in seconds
            ("to", SSINPDataType.FLOAT.value), # end of chunk in seconds
            ("byte", SSINPDataType.INT.value), # number of bytes for the chunk
            ("num", SSINPDataType.INT.value), # number of samples for the chunk
        ]
    )

    def __init__(self, data: np.ndarray, sample_rate: float, chunks: np.ndarray = None, dim_labels: list = None, **kwargs):
        """
        Initialize an SSIStream instance with SSI stream data and metadata.
        """
        super().__init__(data=data, sample_rate=sample_rate, **kwargs)

        # Add Metadata
        if data is not None and chunks is None:
            num_samples = data.shape[0]
            chunks = np.asarray([(0, num_samples / sample_rate, 0, num_samples)], dtype=self.CHUNK_DTYPE)
        ssistream_meta = SSIStreamMetaData(chunks=chunks, dim_labels=dim_labels)
        self.meta_data.expand(ssistream_meta)


class Audio(Stream):
    """
    A class representing an audio data stream.

    This class extends the Stream class with attributes and functionality specific to audio streams.

    Args:
        data (np.ndarray): The audio stream data. Shape is (num_channels, num_samples). Dtype is float.
        sample_rate (float): Sampling rate of the audio stream.
        **kwargs: Additional keyword arguments for Stream.

    """

    def __init__(self, data: np.ndarray, sample_rate: float, **kwargs):
        super().__init__(data=data, sample_rate=sample_rate, **kwargs)


class Video(Stream):
    """
    A class representing video data stream.

    This class extends the Stream class with attributes and functionality specific to video streams.

    Args:
        data (np.ndarray): The video stream data. Shape is (num_samples, height, width, num_channels)
        sample_rate (float): Sampling rate of the video stream.
        **kwargs: Additional keyword arguments for Stream.

    """

    def __init__(self, data: np.ndarray, sample_rate: float, **kwargs):
        super().__init__(data=data, sample_rate=sample_rate, **kwargs)


if __name__ == "__main__":
    # Placeholder for main execution code
    ...



from urllib.request import urlopen
from io import BytesIO
from zipfile import ZipFile
import requests
import shutil
from tqdm.auto import tqdm
from pathlib import Path

def retreive_from_url(url, fp):

    with requests.get(url, stream=True, headers={'Accept-Encoding': None}) as r:

        # check header to get content length, in bytes
        total_length = int(r.headers.get("Content-Length"),0)

        # implement progress bar via tqdm
        with tqdm.wrapattr(r.raw, "read", total=total_length, desc="")as raw:

            # save the output to a file
            with open(fp, 'wb')as output:
                shutil.copyfileobj(raw, output)


def retreive_and_unzip(url, extract_to='./tmp', tmp_dir='./tmp'):
    fn = url.split('/')[-1]
    tmp_dir = Path(tmp_dir)
    tmp_dir.mkdir(parents=True, exist_ok=True)

    tmp_file = tmp_dir / fn

    if tmp_file.is_file():
        print(f'File {tmp_file} already exists. Skipping download.')
    else:
        retreive_from_url(url, tmp_file)

    with ZipFile(tmp_file) as zObject:
        zObject.extractall(path=Path(extract_to))

if __name__ == '__main__':
    retreive_from_url("https://github.com/saveli/syncpy/archive/refs/heads/master.zip",'blub.zip')